# pylint: disable=C0415
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons
import octobot_commons.enums as enums
import octobot_commons.constants as constants
import octobot_commons.logging as logging_util
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_commons.timestamp_util as timestamp_util
import octobot_commons.number_util as number_util

ORDER_TIME_FORMAT = "%m-%d %H:%M"
LOGGER = logging_util.get_logger("PrettyPrinter")


def open_order_pretty_printer(exchange_name, dict_order, markdown=False) -> str:
    """
    Open Order pretty printer
    :param exchange_name: the exchange name
    :param dict_order: the order dict
    :param markdown: if printer use markdown
    :return: the order pretty printed
    """
    try:
        from octobot_trading.enums import (
            ExchangeConstantsOrderColumns,
            TraderOrderType,
            TradeOrderSide,
        )
        from octobot_trading.api.orders import parse_order_type

        _, _, code = get_markers(markdown)
        market = symbol_util.parse_symbol(
            str(dict_order.get(ExchangeConstantsOrderColumns.SYMBOL.value, ""))
        ).quote
        quantity_currency = dict_order.get(
            ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value, ""
        )
        order_type = parse_order_type(dict_order)
        if order_type == TraderOrderType.UNKNOWN:
            order_type = TradeOrderSide(
                dict_order.get(ExchangeConstantsOrderColumns.SIDE.value)
            )
        quantity = dict_order.get(ExchangeConstantsOrderColumns.AMOUNT.value, 0.0)
        price = dict_order.get(ExchangeConstantsOrderColumns.PRICE.value, 0.0)

        return (
            f"{code}{order_type.name.replace('_', ' ')}{code}: {code}"
            f"{get_min_string_from_number(quantity)} "
            f"{quantity_currency}{code} at {code}"
            f"{get_min_string_from_number(price)} {market}{code} "
            f"on {exchange_name.capitalize()}"
        )
    except ImportError:
        LOGGER.error(
            "open_order_pretty_printer requires OctoBot-Trading package installed"
        )
    return ""


def trade_pretty_printer(exchange_name, trade, markdown=False) -> str:
    """
    Trade pretty printer
    :param exchange_name: the exchange name
    :param trade: the trade object
    :param markdown: if printer use markdown
    :return: the trade pretty printed
    """
    try:
        from octobot_trading.enums import TraderOrderType

        _, _, code = get_markers(markdown)
        trade_type = trade.trade_type
        if trade_type == TraderOrderType.UNKNOWN:
            trade_type = trade.side

        trade_executed_time_str = (
            timestamp_util.convert_timestamp_to_datetime(
                trade.executed_time, time_format=ORDER_TIME_FORMAT, local_timezone=True
            )
            if trade.executed_time
            else ""
        )
        return (
            f"{code}{trade_type.name.replace('_', ' ')}{code}: {code}"
            f"{get_min_string_from_number(trade.executed_quantity)} "
            f"{trade.quantity_currency}{code} at {code}"
            f"{get_min_string_from_number(trade.executed_price)} {trade.market}{code} "
            f"{exchange_name.capitalize()} "
            f"{trade_executed_time_str} "
        )
    except ImportError:
        LOGGER.error(
            "open_order_pretty_printer requires OctoBot-Trading package installed"
        )
    return ""


def cryptocurrency_alert(result, final_eval) -> (str, str):
    """
    Cryptocurrency alert
    :param result: the result
    :param final_eval: the final eval
    :return: alert and the markdown alert
    """
    try:
        import telegram.helpers

        _, _, code = get_markers(True)
        display_result = str(result).split(".")[1].replace("_", " ")
        alert = f"Result : {display_result}\n" f"Evaluation : {final_eval}"
        alert_markdown = (
            f"Result : {code}{display_result}{code}\n"
            f"Evaluation : {code}{telegram.helpers.escape_markdown(str(final_eval))}{code}"
        )
        return alert, alert_markdown
    except ImportError:
        LOGGER.error("cryptocurrency_alert requires Telegram package installed")
    return "", ""


def _get_row_pretty_portfolio_row(holdings, currency, ref_market, ref_market_value):
    """
    :return: the portfolio row adapted for a raw format
    """
    str_holdings = get_min_string_from_number(holdings)
    if ref_market:
        return f"{str_holdings} {currency}{get_min_string_from_number(ref_market_value)} {ref_market}"
    return f"{str_holdings} {currency}"


def _get_max_digits(number):
    abs_number = abs(number)
    if abs_number < 0.0001:
        return 8
    if abs_number < 0.01:
        return 6
    if abs_number < 1:
        return 4
    if abs_number < 10000:
        return 2
    return 0


def _get_markdown_pretty_portfolio_row(
    holdings, currency, ref_market, ref_market_value
):
    """
    :return: the portfolio row adapted for a markdown format
    """
    str_currency = "{:<4}".format(currency)
    str_holdings = "{:<12}".format(get_min_string_from_number(holdings))
    str_ref_market_value = "{:<12}".format("")
    if ref_market:
        str_ref_market_value = "{:<12}".format(
            get_min_string_from_number(ref_market_value)
        )
    return f"{str_currency} {str_holdings} {str_ref_market_value}"


def global_portfolio_pretty_print(
    global_portfolio,
    currency_values=None,
    ref_market_name=None,
    separator="\n",
    markdown=False,
) -> str:
    """
    Global portfolio pretty printer
    :param global_portfolio: the global portfolio
    :param currency_values: dict of current currency values {"BTC": 20000, "ETH": 1000 }
    :param ref_market_name: current ref market "USD"
    :param separator: the printer separator
    :param markdown: if printer use markdown
    :return: the global portfolio pretty printed
    """
    results = []
    currency = "currency"
    holdings = "holdings"
    value = "value"
    for asset, asset_dict in global_portfolio.items():
        if asset_dict[constants.PORTFOLIO_TOTAL] > 0:
            holdings_value = 0
            if currency_values and ref_market_name:
                if ref_market_name == asset:
                    holdings_value = asset_dict[constants.PORTFOLIO_TOTAL]
                else:
                    try:
                        holdings_value = (
                            currency_values[asset]
                            * asset_dict[constants.PORTFOLIO_TOTAL]
                        )
                    except KeyError:
                        # no currency value
                        pass
            results.append(
                {
                    currency: asset,
                    holdings: asset_dict[constants.PORTFOLIO_TOTAL],
                    value: holdings_value,
                }
            )
    results.sort(key=lambda r: r[value], reverse=True)
    if markdown:
        # fill lines with empty spaces if necessary
        header = (
            f"{'{:<4}'.format('')} "
            f"{'  {:<9}'.format('Holdings')}  "
            f"{'    {:<7}'.format(ref_market_name or '')}"
        )
        header_separator = f"{'-' * 4}|{'-' * 12}|{'-' * 12}"
        content = separator.join(
            [
                _get_markdown_pretty_portfolio_row(
                    result[holdings], result[currency], ref_market_name, result[value]
                )
                for result in results
            ]
        )
        return f"{header}\n{header_separator}\n{content}"
    return separator.join(
        [
            _get_row_pretty_portfolio_row(
                result[holdings], result[currency], ref_market_name, result[value]
            )
            for result in results
        ]
    )


def portfolio_profitability_pretty_print(
    profitability, profitability_percent, reference
) -> str:
    """
    Profitability pretty printer
    :param profitability: the profitability
    :param profitability_percent: the profitability percent
    :param reference: the reference
    :return: the profitability pretty printed
    """
    difference = (
        f"({get_min_string_from_number(profitability_percent, 5)}%)"
        if profitability_percent is not None
        else ""
    )
    return f"{get_min_string_from_number(profitability, 5)} {reference} {difference}"


def pretty_print_dict(dict_content, default="0", markdown=False) -> str:
    """
    Dict pretty printer
    :param dict_content: the dict to be printed
    :param default: the default printed
    :param markdown: if printer use markdown
    :return: the dict pretty printed
    """
    _, _, code = get_markers(markdown)
    if dict_content:
        result_str = octobot_commons.DICT_BULLET_TOKEN_STR
        return (
            f"{result_str}{code}"
            f"{octobot_commons.DICT_BULLET_TOKEN_STR.join(f'{value} {key}' for key, value in dict_content.items())}"
            f"{code}"
        )
    return default


def round_with_decimal_count(number, max_digits=8) -> float:
    """
    Round a decimal count
    :param number: the number to round
    :param max_digits: the digits
    :return: the rounded number
    """
    if number is None:
        return 0
    return float(get_min_string_from_number(number, max_digits))


def get_min_string_from_number(number, max_digits=None) -> str:
    """
    Get a min string from number
    :param number: the number
    :param max_digits: the mex digits
    :return: the string from number
    """
    max_digits = _get_max_digits(number) if max_digits is None else max_digits
    if number is None or round(number, max_digits) == 0.0:
        return "0"
    if number % 1 != 0:
        number_str = number_util.round_into_str_with_max_digits(number, max_digits)
        # remove post comma trailing 0
        if "." in number_str:
            # remove "0" first and only the "." to avoid removing 2x"0" in 10.0 and returning 1 for example.
            number_str = number_str.rstrip("0").rstrip(".")
        return number_str
    return "{:f}".format(number).split(".")[0]


# return markers for italic, bold and code
def get_markers(markdown=False) -> (str, str, str):
    """
    Get the markdown markers
    :param markdown: if printer use markdown
    :return: the italic marker, the bold marker, the code marker
    """
    if markdown:
        return (
            enums.MarkdownFormat.ITALIC.value,
            enums.MarkdownFormat.BOLD.value,
            enums.MarkdownFormat.CODE.value,
        )
    return "", "", ""



# pylint: disable=R0902, W0703
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import os
import shutil
import uuid
import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.logging as commons_logging
import octobot_commons.json_util as json_util
import octobot_commons.errors as errors


class Profile:
    """
    A profile is managing an OctoBot local configuration: activated tentacles, exchanges, currencies and
    trading settings.
    """

    FULLY_MANAGED_ELEMENTS = [
        constants.CONFIG_CRYPTO_CURRENCIES,
        constants.CONFIG_TRADING,
        constants.CONFIG_TRADER,
        constants.CONFIG_SIMULATOR,
    ]
    PARTIALLY_MANAGED_ELEMENTS = {
        constants.CONFIG_EXCHANGES: {
            constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
            constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
            constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
            constants.CONFIG_ENABLED_OPTION: False,
            constants.CONFIG_EXCHANGE_TYPE: constants.DEFAULT_EXCHANGE_TYPE,
        }
    }
    PARTIALLY_MANAGED_ELEMENTS_FORCED_DEFAULT_KEYS = {
        constants.CONFIG_EXCHANGES: {
            constants.CONFIG_ENABLED_OPTION: False,
        }
    }
    PARTIALLY_MANAGED_ELEMENTS_ALLOWED_KEYS = {
        constants.CONFIG_EXCHANGES: [
            constants.CONFIG_ENABLED_OPTION,
            constants.CONFIG_EXCHANGE_TYPE,
        ]
    }

    def __init__(self, profile_path: str, schema_path: str = None):
        self.profile_id: str = None
        self.path: str = profile_path
        self.schema_path: str = schema_path or constants.PROFILE_FILE_SCHEMA
        self.name: str = None
        self.slug: str = None
        self.description: str = None
        self.avatar: str = None
        self.avatar_path: str = None
        self.origin_url: str = None
        self.auto_update: bool = False
        self.read_only: bool = False
        self.imported: bool = False
        self.complexity: enums.ProfileComplexity = enums.ProfileComplexity.MEDIUM
        self.risk: enums.ProfileRisk = enums.ProfileRisk.MODERATE
        self.profile_type: enums.ProfileType = enums.ProfileType.LIVE
        self.extra_backtesting_time_frames = []

        self.config: dict = {}

    def read_config(self):
        """
        Reads a profile from self.path
        :return: self
        """
        return self.from_dict(json_util.read_file(self.config_file()))

    def from_dict(self, profile_dict: dict):
        """
        Reads a profile from the given dict
        :return: self
        """
        profile_config = profile_dict.get(constants.CONFIG_PROFILE, {})
        self.profile_id = profile_config.get(constants.CONFIG_ID, str(uuid.uuid4()))
        self.name = profile_config.get(constants.CONFIG_NAME, "")
        self.slug = profile_config.get(constants.CONFIG_SLUG, "")
        self.description = profile_config.get(constants.CONFIG_DESCRIPTION, "")
        self.avatar = profile_config.get(constants.CONFIG_AVATAR, "")
        self.origin_url = profile_config.get(constants.CONFIG_ORIGIN_URL, None)
        self.auto_update = profile_config.get(constants.CONFIG_AUTO_UPDATE, False)
        self.read_only = profile_config.get(constants.CONFIG_READ_ONLY, False)
        self.imported = profile_config.get(constants.CONFIG_IMPORTED, False)
        self.complexity = enums.ProfileComplexity(
            profile_config.get(
                constants.CONFIG_COMPLEXITY, enums.ProfileComplexity.MEDIUM.value
            )
        )
        self.risk = enums.ProfileRisk(
            profile_config.get(constants.CONFIG_RISK, enums.ProfileRisk.MODERATE.value)
        )
        self.profile_type = enums.ProfileType(
            profile_config.get(constants.CONFIG_TYPE, enums.ProfileType.LIVE.value)
        )
        self.extra_backtesting_time_frames = profile_config.get(
            constants.CONFIG_EXTRA_BACKTESTING_TIME_FRAMES, []
        )
        self.config = profile_dict[constants.PROFILE_CONFIG]
        if self.avatar and self.path:
            avatar_path = os.path.join(self.path, self.avatar)
            if os.path.isfile(avatar_path):
                self.avatar_path = avatar_path
        return self

    def save_config(self, global_config: dict):
        """
        Save this profile config
        :param global_config: the bot config containing profile data
        :return: None
        """
        for element in self.FULLY_MANAGED_ELEMENTS:
            if element in global_config:
                self.config[element] = global_config[element]
        self.sync_partially_managed_elements(global_config)
        self.validate_and_save_config()

    def remove_deleted_elements(self, global_config):
        """
        Removes elements from self.PARTIALLY_MANAGED_ELEMENTS
        that are in profile but not in global config
        """
        for element in self.PARTIALLY_MANAGED_ELEMENTS:
            if element in global_config and element in self.config:
                current_elements = list(self.config[element])
                to_keep_elements = set(global_config[element])
                for key in current_elements:
                    if key not in to_keep_elements:
                        self.config[element].pop(key)

    def sync_partially_managed_elements(self, global_config):
        """
        Update the partially managed elements of this profile using the given configuration
        """
        for element in self.PARTIALLY_MANAGED_ELEMENTS:
            if element in global_config:
                allowed_keys = self.PARTIALLY_MANAGED_ELEMENTS_ALLOWED_KEYS.get(
                    element, None
                )
                if allowed_keys is not None:
                    self._filter_fill_elements(
                        global_config, self.config, element, allowed_keys
                    )

    def validate(self):
        """
        Validate this profile configuration against self.schema_path
        :return:
        """
        try:
            json_util.validate(self.as_dict(), self.schema_path)
        except FileNotFoundError as err:
            commons_logging.get_logger("ProfileSaver").warning(
                f"Impossible to validate profile: {err} ({err.__class__.__name__})"
            )

    def validate_and_save_config(self) -> None:
        """
        JSON validates this profile and then saves its configuration file
        :return: None
        """
        self.validate()
        self.save()

    def save(self) -> None:
        """
        Saves the current profile configuration file
        :return: None
        """
        json_util.safe_dump(self.as_dict(), self.config_file())

    def rename_folder(self, new_name, should_raise) -> str:
        """
        rename the profile folder
        :param new_name: name of the new folder
        :param should_raise: raises ProfileConflictError if the profile can't be renamed
        :return: the new profile path
        """
        new_path = os.path.join(os.path.split(self.path)[0], new_name)
        if os.path.exists(new_path):
            if should_raise:
                raise errors.ProfileConflictError(
                    "Skipping folder renaming: a profile already exists at this path"
                )
            return self.path
        try:
            os.rename(self.path, new_path)
            self.path = new_path
        except Exception as err:
            commons_logging.get_logger("ProfileRenamer").error(
                f"Error when renaming profile: {err}"
            )
            raise errors.ProfileConflictError from err
        return self.path

    def duplicate(self, name: str = None, description: str = None):
        """
        Duplicates the current profile and associates it with a new profile_id
        :param name: name of the profile to create, uses the original's one by default
        :param description: description of the profile to create, uses the original's one by default
        :return: the created profile
        """
        clone = copy.deepcopy(self)
        clone.name = name or clone.name
        clone.description = description or clone.description
        clone.profile_id = str(uuid.uuid4())
        clone.read_only = False
        clone.imported = False
        clone.origin_url = None
        clone.auto_update = False
        try:
            clone.path = os.path.join(
                os.path.split(self.path)[0], f"{clone.name}_{clone.profile_id}"
            )
            shutil.copytree(self.path, clone.path)
        except OSError:
            # invalid profile name for a filename
            clone.path = os.path.join(os.path.split(self.path)[0], clone.profile_id)
            shutil.copytree(self.path, clone.path)
        clone.save()
        return clone

    def get_tentacles_config_path(self) -> str:
        """
        :return: The tentacles configurations path
        """
        return os.path.join(self.path, constants.CONFIG_TENTACLES_FILE)

    def as_dict(self) -> dict:
        """
        :return: A dict representation of this profile configuration
        """
        return {
            constants.CONFIG_PROFILE: {
                constants.CONFIG_ID: self.profile_id,
                constants.CONFIG_NAME: self.name,
                constants.CONFIG_SLUG: self.slug,
                constants.CONFIG_DESCRIPTION: self.description,
                constants.CONFIG_AVATAR: self.avatar,
                constants.CONFIG_ORIGIN_URL: self.origin_url,
                constants.CONFIG_AUTO_UPDATE: self.auto_update,
                constants.CONFIG_READ_ONLY: self.read_only,
                constants.CONFIG_IMPORTED: self.imported,
                constants.CONFIG_COMPLEXITY: (
                    self.complexity.value if self.complexity else None
                ),
                constants.CONFIG_RISK: self.risk.value if self.risk else None,
                constants.CONFIG_TYPE: (
                    self.profile_type.value if self.profile_type else None
                ),
                constants.CONFIG_EXTRA_BACKTESTING_TIME_FRAMES: self.extra_backtesting_time_frames,
            },
            constants.PROFILE_CONFIG: self.config,
        }

    def config_file(self):
        """
        :return: the path to this profile config file
        """
        return os.path.join(self.path, constants.PROFILE_CONFIG_FILE)

    def merge_partially_managed_element_into_config(self, config: dict, element: str):
        """
        Merge this profile configuration's partially managed element into the given config
        :param config: dict to merge this profile configuration's partially managed element into
        :param element: the partially managed element to merge
        :return: None
        """
        Profile._merge_partially_managed_element(
            config, self.config, element, Profile.PARTIALLY_MANAGED_ELEMENTS[element]
        )

    @staticmethod
    def _merge_partially_managed_element(
        config: dict, profile_config: dict, element: str, template: dict
    ):
        if element in config:
            Profile._merge_profile_values(config, profile_config, element, template)
            Profile._apply_forced_default_values(config, profile_config, element)
        else:
            # use profile value for element
            config[element] = {
                key: Profile._get_element_from_template(template, val)
                for key, val in profile_config[element].items()
            }

    @staticmethod
    def _merge_profile_values(
        config: dict, profile_config: dict, element: str, template: dict
    ):
        for key, val in profile_config[element].items():
            if key in config[element]:
                if isinstance(config[element][key], dict):
                    # merge profile values for element[key]
                    Profile._merge_partially_managed_element(
                        config[element], profile_config[element], key, template
                    )
                else:
                    # overwrite element[key] by profile value
                    config[element][key] = copy.deepcopy(profile_config[element][key])
            else:
                # use profile value for element[key]
                if isinstance(val, dict):
                    config[element][key] = Profile._get_element_from_template(
                        template, val
                    )
                else:
                    config[element][key] = val

    @staticmethod
    def _apply_forced_default_values(config: dict, profile_config: dict, element: str):
        if element in Profile.PARTIALLY_MANAGED_ELEMENTS_FORCED_DEFAULT_KEYS:
            for config_key, config_val in config[element].items():
                if config_key not in profile_config[element]:
                    for config_sub_element in config_val:
                        if (
                            config_sub_element
                            in Profile.PARTIALLY_MANAGED_ELEMENTS_FORCED_DEFAULT_KEYS[
                                element
                            ]
                        ):
                            # item not in profile, it will be added to profile upon save
                            # use forced default profile value for forced default keys
                            config[element][config_key][config_sub_element] = (
                                Profile.PARTIALLY_MANAGED_ELEMENTS_FORCED_DEFAULT_KEYS[
                                    element
                                ][config_sub_element]
                            )

    @staticmethod
    def _get_element_from_template(template: dict, profile_values: dict) -> dict:
        merged_values = copy.deepcopy(template)
        merged_values.update(profile_values)
        return merged_values

    @staticmethod
    def _filter_fill_elements(
        config: dict, profile_config: dict, element: str, allowed_keys: list
    ):
        if element in config:
            # reset profile element to avoid saving outdated data
            profile_config[element] = {}
            for key, value in config[element].items():
                if isinstance(value, dict):
                    # handle nested elements
                    Profile._filter_fill_elements(
                        config[element], profile_config[element], key, allowed_keys
                    )
                else:
                    # save allowed keys
                    if key in allowed_keys:
                        profile_config[element][key] = value

    @staticmethod
    def load_profile(profiles_path, profile_id, schema_path: str = None):
        """
        :param profiles_path: the path to look for the profile
        :param profile_id: the required profile id
        :return: the loaded profile
        """
        for profile in Profile.get_all_profiles(profiles_path, schema_path=schema_path):
            if profile.profile_id == profile_id:
                return profile
        raise errors.NoProfileError(f"No profile with id: {profile_id}")

    @staticmethod
    def get_all_profiles(profiles_path, ignore: str = None, schema_path: str = None):
        """
        Loads profiles found in the given directory
        :param profiles_path: Path to a directory containing profiles
        :param ignore: A profile path to ignore
        :param schema_path: Path to the json schema to pass to the created profile instances
        :return: the profile instances list
        """
        profiles = []
        ignored_path = None if ignore is None else os.path.normpath(ignore)
        for profile_entry in os.scandir(profiles_path):
            if (
                ignored_path is None
                or os.path.normpath(profile_entry.path) != ignored_path
            ):
                profile = Profile._load_profile(profile_entry.path, schema_path)
                if profile is not None:
                    profiles.append(profile)
        return profiles

    @staticmethod
    def _load_profile(profile_path: str, schema_path: str):
        logger = commons_logging.get_logger("ProfileExplorer")
        profile = Profile(profile_path, schema_path)
        try:
            if os.path.isfile(profile.config_file()):
                profile.read_config()
                return profile
            logger.debug(
                f"Ignored {profile_path} as it does not contain a profile configuration"
            )
        except Exception as err:
            logger.exception(
                err,
                True,
                f"Ignored profile due to an error upon reading '{profile_path}': {err}",
            )
        return None

    @staticmethod
    def get_all_profiles_ids(profiles_path, ignore: str = None):
        """
        Get ids of profiles found in the given directory
        :param profiles_path: Path to a directory containing profiles
        :param ignore: A profile path to ignore in ids listing
        :return: the profile ids list
        """
        return [
            profile.profile_id
            for profile in Profile.get_all_profiles(profiles_path, ignore)
        ]



# pylint: disable=R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from typing import Tuple, Optional
from cryptography.exceptions import InvalidSignature

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes, serialization


def generate_ecdsa_key_pair(
    curve: Optional[ec.EllipticCurve] = None,
) -> Tuple[bytes, bytes]:
    """Generate an ECDSA key pair

    This function generates a private and public ECDSA key pair equivalent to:
    - openssl ecparam -name secp256r1 -genkey -out user_ecdsa_private_key.pem
    - openssl ec -in user_ecdsa_private_key.pem -pubout -out user_ecdsa_public_key.pem

    :param curve: The elliptic curve to use. Defaults to SECP256R1 (secp256r1/prime256v1).
    :type curve: Optional[ec.EllipticCurve]
    :return: A tuple containing (private_key_pem, public_key_pem) as bytes.
    :rtype: Tuple[bytes, bytes]
    """
    if curve is None:
        curve = ec.SECP256R1()

    # Generate private key
    private_key = ec.generate_private_key(curve)

    # Serialize private key to PEM format
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    # Extract and serialize public key to PEM format
    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return private_key_pem, public_key_pem


def sign_data(data: bytes, private_key_pem: bytes) -> bytes:
    """Generate an ECDSA signature for data

    This function generates a signature equivalent to:
    - openssl dgst -sha256 -sign ecdsa_private_key.pem -out signature.bin data.bin

    :param data: The data to sign.
    :type data: bytes
    :param private_key_pem: The ECDSA private key in PEM format.
    :type private_key_pem: bytes
    :return: The signature as bytes.
    :rtype: bytes
    :raises ValueError: If the private key cannot be loaded or is invalid.
    """
    # Load the private key from PEM format
    private_key = serialization.load_pem_private_key(
        private_key_pem,
        password=None,
    )

    if not isinstance(private_key, ec.EllipticCurvePrivateKey):
        raise ValueError("Private key must be an ECDSA key.")

    # Hash the data with SHA-256 and sign it
    signature = private_key.sign(data, ec.ECDSA(hashes.SHA256()))

    return signature


def verify_signature(data: bytes, public_key_pem: bytes, signature: bytes) -> bool:
    """Verify an ECDSA signature for data

    This function verifies a signature equivalent to:
    - openssl dgst -sha256 -verify ecdsa_public_key.pem -signature signature.bin data.bin

    :param data: The original data that was signed.
    :type data: bytes
    :param public_key_pem: The ECDSA public key in PEM format.
    :type public_key_pem: bytes
    :param signature: The signature to verify.
    :type signature: bytes
    :return: True if the signature is valid, False otherwise.
    :rtype: bool
    :raises ValueError: If the public key cannot be loaded or is invalid.
    """
    # Load the public key from PEM format
    public_key = serialization.load_pem_public_key(public_key_pem)

    if not isinstance(public_key, ec.EllipticCurvePublicKey):
        raise ValueError("Public key must be an ECDSA key.")

    try:
        # Hash the data with SHA-256 and verify the signature
        public_key.verify(signature, data, ec.ECDSA(hashes.SHA256()))
        return True
    except InvalidSignature:
        return False



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import numpy as np

from octobot_commons.constants import START_PENDING_EVAL_NOTE, INIT_EVAL_NOTE
from octobot_commons.evaluators_util import check_valid_eval_note


def test_check_valid_eval_note():
    assert not check_valid_eval_note(START_PENDING_EVAL_NOTE)
    assert check_valid_eval_note(True)
    assert check_valid_eval_note({"a": 1})
    assert check_valid_eval_note({"a": 1}, eval_time=1, expiry_delay=2, current_time=1)
    assert check_valid_eval_note({"a": 1}, eval_time=1, expiry_delay=2, current_time=2)
    assert not check_valid_eval_note({"a": 1}, eval_time=1, expiry_delay=2, current_time=3)

    assert check_valid_eval_note(INIT_EVAL_NOTE)
    # UNSET_EVAL_TYPE




#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.cryptography import encryption
from octobot_commons.cryptography.encryption import (
    generate_rsa_key_pair,
    generate_aes_key,
    generate_iv,
    rsa_decrypt_aes_key,
    rsa_encrypt_aes_key,
    aes_gcm_decrypt,
    aes_gcm_encrypt,
    pbkdf2_derive_key_from_pin,
    pbkdf2_encrypt_aes_key,
    pbkdf2_decrypt_aes_key,
)
from octobot_commons.cryptography import signing
from octobot_commons.cryptography.signing import (
    generate_ecdsa_key_pair,
    sign_data,
    verify_signature,
)


__all__ = [
    "generate_rsa_key_pair",
    "generate_aes_key",
    "generate_iv",
    "rsa_decrypt_aes_key",
    "rsa_encrypt_aes_key",
    "aes_gcm_decrypt",
    "aes_gcm_encrypt",
    "pbkdf2_derive_key_from_pin",
    "pbkdf2_encrypt_aes_key",
    "pbkdf2_decrypt_aes_key",
    "generate_ecdsa_key_pair",
    "sign_data",
    "verify_signature",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import functools

import octobot_commons.dsl_interpreter
import octobot_commons.tentacles_management
import octobot_commons.constants


@functools.lru_cache(maxsize=16)
def get_all_operators(
    *libraries: str,
) -> list["octobot_commons.dsl_interpreter.Operator"]:
    """
    Get all operators from the DSL interpreter.
    This function is cached and will return the same list of operators every time.
    All operators must be subclasses of octobot_commons.dsl_interpreter.Operator
    All operators must have been imported before calling this function for the first time.
    :param libraries: List of libraries to filter operators by. If None, all operators will be returned.
    By default, operators are in the octobot_commons.constants.BASE_OPERATORS_LIBRARY library.
    """
    libraries_filter = set(libraries) if libraries else None
    all_with_abstracts = (
        operator
        for operator in octobot_commons.tentacles_management.get_all_classes_from_parent(
            octobot_commons.dsl_interpreter.Operator
        )
        if (
            libraries_filter is None
            or operator.get_library()
            in libraries_filter  # pylint: disable=unsupported-membership-test
        )
        # contextual operators should not be included by default
        and operator.get_library()
        != octobot_commons.constants.CONTEXTUAL_OPERATORS_LIBRARY
    )
    non_abstract_operators = []
    for operator in all_with_abstracts:
        try:
            operator.get_name()
            non_abstract_operators.append(operator)
        except NotImplementedError:
            # this is an abstract operator
            pass
    return non_abstract_operators


def clear_get_all_operators_cache():
    """
    Clear the cache of the get_all_operators function.
    """
    get_all_operators.cache_clear()



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.errors as errors
import octobot_commons.dict_util as dict_util


class GenericDatabaseCache:
    MAX_CACHE_SIZE = 512

    def __init__(self):
        # used for select / contains methods
        self.rows_cache = {}
        # used for cached_query() (used in upsert)
        self.query_cache = {}
        # used for cached_uuid() (used in upsert)
        self.uuid_cache = {}

    def register(self, table, row, result=None, uuid=None):
        """
        Saves the given row/query in local cache
        :param table: the associated table
        :param row: the row/query to save
        :param result: select result to save
        :param uuid: uuid to save
        """
        cached = False
        try:
            if uuid is not None:
                cached = True
                try:
                    self.uuid_cache[table][row] = uuid
                except KeyError:
                    self.uuid_cache[table] = {row: uuid}
            elif result is not None:
                cached = True
                try:
                    self.query_cache[table][row] = result
                except KeyError:
                    self.query_cache[table] = {row: result}
        except TypeError as err:
            # might happen when row can't be hashed: impossible to cache it in this case
            raise errors.UncachableValue(f"Unhashable row: {row}") from err
        if not cached:
            self._add_to_rows_cache(table, row)

    def _add_to_rows_cache(self, table, row):
        try:
            if len(self.rows_cache[table]) >= self.MAX_CACHE_SIZE:
                self.rows_cache[table] = self.rows_cache[table][
                    self.MAX_CACHE_SIZE // 2 :
                ]
            self.rows_cache[table].append(row)
        except KeyError:
            self.rows_cache[table] = [row]

    def has(self, table):
        """
        :param table: table name
        :return: True if the given table is in rows_cache
        """
        return table in self.rows_cache

    def cached_uuid(self, table, identifier):
        """
        :param table: table name
        :param identifier: identifier of to look for
        :return: the cached uuid of the given identifier
        """
        try:
            return self.uuid_cache[table][identifier]
        except KeyError:
            return None

    def cached_query(self, table, identifier):
        """
        :param table: table name
        :param identifier: identifier of to look for
        :return: the cached query of the given identifier
        """
        try:
            return self.query_cache[table][identifier]
        except KeyError:
            return None

    def contains_row(self, table, val_by_keys):
        """
        :param table: table name
        :param val_by_keys: dict to look for
        :return: True if a row of the local cache contains every value of the given dict
        """
        # Should check the real database in case this returns false
        try:
            for element in self.rows_cache[table]:
                if dict_util.contains_each_element(element, val_by_keys):
                    return True
        except KeyError:
            pass
        return False

    def delete_from_rows_cache(self, table, val_by_keys):
        """
        :param table: table name
        :param val_by_keys: dict to look for
        :return: True if a row of the local cache contains every value of the given dict
        """
        # Should check the real database in case this returns false
        try:
            self.rows_cache[table] = [
                element
                for element in self.rows_cache[table]
                if not dict_util.contains_each_element(element, val_by_keys)
            ]
        except KeyError:
            pass

    def clear(self, table=None):
        """
        Resets the current cache
        """
        if table:
            self.rows_cache.pop(table, None)
            self.query_cache.pop(table, None)
            self.uuid_cache.pop(table, None)
        else:
            self.rows_cache = {}
            self.query_cache = {}
            self.uuid_cache = {}



# pylint: disable=R0913
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import functools
import typing
import collections

import octobot_commons
import octobot_commons.constants as constants
import octobot_commons.symbols.symbol


_MAX_PARSED_SYMBOLS_CACHE_SIZE = 2048


@functools.lru_cache(maxsize=_MAX_PARSED_SYMBOLS_CACHE_SIZE)
def parse_symbol(symbol: str):
    """
    Parse the specified symbol into a Symbol object
    :param symbol: the symbol to parse
    :return: Symbol object
    """
    return octobot_commons.symbols.symbol.Symbol(symbol)


def merge_symbol(symbol: str) -> str:
    """
    Return merged currency and market without /
    :param symbol: the specified symbol
    :return: merged currency and market without /
    """
    return symbol.replace(octobot_commons.MARKET_SEPARATOR, "").replace(
        octobot_commons.SETTLEMENT_ASSET_SEPARATOR, "_"
    )


def is_symbol(value: str, separator: str = octobot_commons.MARKET_SEPARATOR) -> bool:
    """
    Check if the given string is a symbol or a coin based on the separator
    :param value: the string to check
    :param separator: the separator to use for checking (e.g., "/" for "BTC/USDT")
    :return: True if the string is a symbol (contains the separator), False if it's a coin (single currency)
    """
    return separator in value


def merge_currencies(
    currency: str,
    market: str,
    settlement_asset: typing.Optional[str] = None,
    identifier: typing.Optional[str] = None,
    strike_price: typing.Optional[str] = None,
    option_type: typing.Optional[str] = None,
    market_separator: str = octobot_commons.MARKET_SEPARATOR,
    settlement_separator: str = octobot_commons.SETTLEMENT_ASSET_SEPARATOR,
    option_separator: str = octobot_commons.OPTION_SEPARATOR,
) -> str:
    """
    Merge currency and market
    :param currency: the base currency
    :param market: the quote currency
    :param settlement_asset: the settlement asset currency (unused for spot trading)
    :param strike_price: the strike price or time (used for options)
    :param market_separator: the separator between currency and market
    :param settlement_separator: the separator between the pair and reference market
    :param option_separator: the separator between the option details
    :return: currency and market merged
    """
    symbol = octobot_commons.symbols.symbol.Symbol(
        f"{currency}{market_separator}{market}", market_separator=market_separator
    )
    if settlement_asset is not None:
        symbol.settlement_asset = settlement_asset
    if strike_price:
        symbol.strike_price = strike_price
    if identifier:
        symbol.identifier = identifier
    if option_type:
        symbol.option_type = option_type
    return symbol.merged_str_symbol(
        market_separator=market_separator,
        settlement_separator=settlement_separator,
        option_separator=option_separator,
    )


def convert_symbol(
    symbol: str,
    symbol_separator: str,
    new_symbol_separator: str = octobot_commons.MARKET_SEPARATOR,
    should_uppercase: bool = False,
    should_lowercase: bool = False,
    base_and_quote_only: bool = False,
) -> str:
    """
    Convert symbol according to parameter
    :param symbol: the symbol to convert
    :param symbol_separator: the symbol separator
    :param new_symbol_separator: the new symbol separator
    :param should_uppercase: if it should be concerted to uppercase
    :param should_lowercase: if it should be concerted to lowercase
    :param base_and_quote_only: if it should only contain base and quote from the given symbol
    :return:
    """
    if base_and_quote_only:
        symbol = symbol.split(octobot_commons.SETTLEMENT_ASSET_SEPARATOR)[0]
    if should_uppercase:
        return symbol.replace(symbol_separator, new_symbol_separator).upper()
    if should_lowercase:
        return symbol.replace(symbol_separator, new_symbol_separator).lower()
    return symbol.replace(symbol_separator, new_symbol_separator)


def is_usd_like_coin(coin: str) -> bool:
    """
    :return: True if the given coin is a USD-like coin
    """
    return coin in constants.USD_LIKE_COINS


def get_most_common_usd_like_symbol(pairs: list[str]) -> str:
    """
    :return: The most common USD like symbol from the given pairs
    """
    if not pairs:
        raise ValueError("Pairs cannot be empty")
    symbols = []
    for pair in pairs:
        parsed = octobot_commons.symbols.symbol.Symbol(pair)
        symbols.append(parsed.quote)
        symbols.append(parsed.base)
    counter = collections.Counter(symbols)
    for symbol, _ in counter.most_common():
        if is_usd_like_coin(symbol):
            return symbol
    raise ValueError("Pairs cannot be empty")



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import os
import time

from mock import AsyncMock, patch
import pytest

from octobot_commons.async_job import AsyncJob
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle

pytestmark = pytest.mark.asyncio


async def callback():
    pass


async def test_has_enough_time_elapsed():
    job = AsyncJob(callback)
    if not os.getenv('CYTHON_IGNORE'):
        assert job._has_enough_time_elapsed()
        job.last_execution_time = time.time()
        assert job._has_enough_time_elapsed()
    job.stop()


async def test_has_enough_time_elapsed_with_delays():
    job = AsyncJob(callback, min_execution_delay=0.2)
    if not os.getenv('CYTHON_IGNORE'):
        assert job._has_enough_time_elapsed()
        job.last_execution_time = time.time()
        assert not job._has_enough_time_elapsed()
        await asyncio.sleep(0.2)
        assert job._has_enough_time_elapsed()
    job.stop()


async def test_should_run_with_dependencies():
    job2 = AsyncJob(callback)
    job3 = AsyncJob(callback)
    job = AsyncJob(callback)
    job.add_job_dependency(job2)
    job.add_job_dependency(job3)
    if not os.getenv('CYTHON_IGNORE'):
        assert job._should_run_job()
        job2.idle_task_event.clear()
        assert not job._are_job_dependencies_idle()
        job2.idle_task_event.clear()
        job3.idle_task_event.clear()
        assert not job._are_job_dependencies_idle()
        job2.idle_task_event.set()
        assert not job._are_job_dependencies_idle()
        job3.idle_task_event.set()
        assert job._are_job_dependencies_idle()
    job.stop()


async def test_first_execution_delay():
    callback_mock = AsyncMock(__name__="callback_mock")
    job = AsyncJob(callback_mock, first_execution_delay=0.4, execution_interval_delay=5, min_execution_delay=5)
    await wait_asyncio_next_cycle()
    callback_mock.assert_not_called()
    await job.run()
    await wait_asyncio_next_cycle()
    await asyncio.sleep(0.1)
    callback_mock.assert_not_called()
    await asyncio.sleep(0.4)
    callback_mock.assert_called_once()
    job.stop()


async def test_clear():
    job = AsyncJob(callback)
    job.clear()
    if not os.getenv('CYTHON_IGNORE'):
        assert not job.job_dependencies
        assert not job.job_task
        assert not job.job_periodic_task


async def test_run_stop_run():
    job = AsyncJob(callback, execution_interval_delay=0.5, min_execution_delay=0.2)
    if not os.getenv('CYTHON_IGNORE'):
        with patch.object(job, 'callback', new=AsyncMock()) as mocked_test_job_callback:
            await wait_asyncio_next_cycle()
            mocked_test_job_callback.assert_not_called()
            assert not job.is_stopped()
            assert not job.is_started
            await job.run()
            await wait_asyncio_next_cycle()
            mocked_test_job_callback.assert_called_once()
            assert job.is_started
            assert not job.is_stopped()
            job.stop()
            assert job.is_stopped()
            await wait_asyncio_next_cycle()
            assert not job.is_started
            mocked_test_job_callback.assert_called_once()
            await job.run()
            assert not job.is_stopped()
            await asyncio.sleep(0.7)
            assert job.is_started
            assert mocked_test_job_callback.call_count == 2

            await asyncio.sleep(0.5)
            assert mocked_test_job_callback.call_count in (3, 4)    # can be 3 or 4 depending on the current computer
            job.stop()
            assert job.is_stopped()


async def test_run():
    job = AsyncJob(callback, execution_interval_delay=0.5, min_execution_delay=0.2)
    if not os.getenv('CYTHON_IGNORE'):
        with patch.object(job, 'callback', new=AsyncMock()) as mocked_test_job_callback:
            await wait_asyncio_next_cycle()
            mocked_test_job_callback.assert_not_called()
            assert not job.is_started
            await job.run()
            await wait_asyncio_next_cycle()
            mocked_test_job_callback.assert_called_once()
            assert job.is_started

            # delay has not been waited
            await job.run()
            await wait_asyncio_next_cycle()
            mocked_test_job_callback.assert_called_once()
            assert job.is_started

            await asyncio.sleep(0.1)
            mocked_test_job_callback.assert_called_once()

            await asyncio.sleep(0.6)
            assert mocked_test_job_callback.call_count == 2

            await job.run(force=True, wait_for_task_execution=True)
            assert mocked_test_job_callback.call_count == 3

            await wait_asyncio_next_cycle()
            # no periodic trigger yet
            assert mocked_test_job_callback.call_count == 3

            await job.run(force=True, wait_for_task_execution=False)
            # task not yet executed
            assert mocked_test_job_callback.call_count == 3
            await wait_asyncio_next_cycle()
            assert mocked_test_job_callback.call_count == 4

            await asyncio.sleep(0.7)
            # periodic auto trigger
            assert mocked_test_job_callback.call_count in (5, 6)    # can be 5 or 6 depending on the current computer

    job.stop()



# pylint: disable=C0103
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum


class TimeFrames(enum.Enum):
    """
    OctoBot supported time frames values
    """

    ONE_MINUTE = "1m"
    THREE_MINUTES = "3m"
    FIVE_MINUTES = "5m"
    FIFTEEN_MINUTES = "15m"
    THIRTY_MINUTES = "30m"
    ONE_HOUR = "1h"
    TWO_HOURS = "2h"
    THREE_HOURS = "3h"
    FOUR_HOURS = "4h"
    SIX_HOURS = "6h"
    HEIGHT_HOURS = "8h"
    TWELVE_HOURS = "12h"
    ONE_DAY = "1d"
    THREE_DAYS = "3d"
    ONE_WEEK = "1w"
    ONE_MONTH = "1M"
    ONE_YEAR = "1y"


TimeFramesMinutes = {
    TimeFrames.ONE_MINUTE: 1,
    TimeFrames.THREE_MINUTES: 3,
    TimeFrames.FIVE_MINUTES: 5,
    TimeFrames.FIFTEEN_MINUTES: 15,
    TimeFrames.THIRTY_MINUTES: 30,
    TimeFrames.ONE_HOUR: 60,
    TimeFrames.TWO_HOURS: 120,
    TimeFrames.THREE_HOURS: 180,
    TimeFrames.FOUR_HOURS: 240,
    TimeFrames.SIX_HOURS: 360,
    TimeFrames.HEIGHT_HOURS: 480,
    TimeFrames.TWELVE_HOURS: 720,
    TimeFrames.ONE_DAY: 1440,
    TimeFrames.THREE_DAYS: 4320,
    TimeFrames.ONE_WEEK: 10080,
    TimeFrames.ONE_MONTH: 43200,
    TimeFrames.ONE_YEAR: 524160,
}


class PriceIndexes(enum.Enum):
    """
    Default candle price index correspondence
    """

    IND_PRICE_TIME = 0
    IND_PRICE_OPEN = 1
    IND_PRICE_HIGH = 2
    IND_PRICE_LOW = 3
    IND_PRICE_CLOSE = 4
    IND_PRICE_VOL = 5


class PriceStrings(enum.Enum):
    """
    Default candle price str
    """

    STR_PRICE_TIME = "time"
    STR_PRICE_CLOSE = "close"
    STR_PRICE_OPEN = "open"
    STR_PRICE_HIGH = "high"
    STR_PRICE_LOW = "low"
    STR_PRICE_VOL = "vol"


class OptionTypes(enum.Enum):
    """
    Default option type
    """

    PUT = "P"
    CALL = "C"


class PlatformsName(enum.Enum):
    """
    OctoBot supported platforms name
    """

    WINDOWS = "nt"
    LINUX = "posix"
    MAC = "mac"


class OctoBotTypes(enum.Enum):
    """
    OctoBot running types
    """

    BINARY = "binary"
    PYTHON = "python"
    DOCKER = "docker"


class MarkdownFormat(enum.Enum):
    """
    Markdown formating
    """

    ITALIC = "_"
    BOLD = "*"
    CODE = "`"
    IGNORE = 1
    NONE = 0


class OctoBotChannelSubjects(enum.Enum):
    """
    OctoBot Channel subjects
    """

    NOTIFICATION = "notification"
    CREATION = "creation"
    UPDATE = "update"
    DELETION = "deletion"
    ERROR = "error"


class UserCommands(enum.Enum):
    """
    Allowed user commands
    """

    MANUAL_TRIGGER = "manual_trigger"
    OPTIMIZE_INITIAL_PORTFOLIO = "optimize_initial_portfolio"
    TRIGGER_HEALTH_CHECK = "trigger_health_check"
    RELOAD_CONFIG = "reload_config"
    RELOAD_SCRIPT = "reload_script"
    CLEAR_PLOTTING_CACHE = "clear_plotting_cache"
    CLEAR_SIMULATED_ORDERS_CACHE = "clear_simulated_orders_cache"
    CLEAR_SIMULATED_TRADES_CACHE = "clear_simulated_trades_cache"
    CLEAR_SIMULATED_TRANSACTIONS_CACHE = "clear_simulated_transactions_cache"


class MultiprocessingLocks(enum.Enum):
    """
    Keys to multiprocessing lock
    """

    DBLock = "db_lock"


class CacheDatabaseTables(enum.Enum):
    """
    Tables in cache databases
    """

    CACHE = "cache"
    METADATA = "metadata"


class CacheDatabaseColumns(enum.Enum):
    """
    Keys/columns in cache databases tables
    """

    TIMESTAMP = "t"
    VALUE = "v"
    TYPE = "ty"
    TRIGGERED_AFTER_CANDLES_CLOSE = "triggered_after_candles_close"


class PlotAttributes(enum.Enum):
    KIND = "kind"
    X = "x"
    Y = "y"
    Z = "z"
    OPEN = "open"
    HIGH = "high"
    LOW = "low"
    CLOSE = "close"
    VOLUME = "volume"
    TITLE = "title"
    TEXT = "text"
    SUB_ELEMENTS = "sub_elements"
    ELEMENTS = "elements"
    NAME = "name"
    DATA = "data"
    X_TYPE = "x_type"
    Y_TYPE = "y_type"
    MODE = "mode"
    LINE_SHAPE = "line_shape"
    OWN_XAXIS = "own_xaxis"
    OWN_YAXIS = "own_yaxis"
    SIDE = "side"
    VALUE = "value"
    CONFIG = "config"
    SCHEMA = "schema"
    TENTACLE = "tentacle"
    TENTACLE_TYPE = "tentacle_type"
    COLUMNS = "columns"
    ROWS = "rows"
    SEARCHES = "searches"
    IS_HIDDEN = "is_hidden"
    TYPE = "type"
    COLOR = "color"
    HTML = "html"
    SIZE = "size"
    SHAPE = "shape"
    SYMBOL = "symbol"


class BacktestingMetadata(enum.Enum):
    ID = "id"
    GAINS = "gains"
    PERCENT_GAINS = "% gains"
    MARKETS_PROFITABILITY = "markets profitability"
    END_PORTFOLIO = "end portfolio"
    START_PORTFOLIO = "start portfolio"
    WIN_RATE = "% win rate"
    DRAW_DOWN = "% draw down"
    COEFFICIENT_OF_DETERMINATION_MAX_BALANCE = "R² max balance"
    COEFFICIENT_OF_DETERMINATION_END_BALANCE = "R² end balance"
    SYMBOLS = "symbols"
    TIME_FRAMES = "time frames"
    START_TIME = "start time"
    END_TIME = "end time"
    DURATION = "duration"
    ENTRIES = "entries"
    WINS = "wins"
    LOSES = "loses"
    TRADES = "trades"
    TIMESTAMP = "timestamp"
    NAME = "name"
    LEVERAGE = "leverage"
    OPTIMIZATION_CAMPAIGN = "optimization campaign"
    USER_INPUTS = "user inputs"
    BACKTESTING_FILES = "backtesting files"
    CHILDREN = "children"
    OPTIMIZER_ID = "optimizer id"
    EXCHANGE = "exchange"


class DBRows(enum.Enum):
    ID = "id"
    REFERENCE_MARKET = "ref_market"
    EXCHANGE = "exchange"
    EXCHANGES = "exchanges"
    FUTURE_CONTRACTS = "future_contracts"
    PAIR = "pair"
    TIME_FRAME = "time_frame"
    VALUE = "value"
    START_TIME = "start_time"
    END_TIME = "end_time"
    TRADING_TYPE = "trading_type"
    TRADING_MODE = "trading_mode"
    SYMBOL = "symbol"
    SYMBOLS = "symbols"
    FEES_AMOUNT = "fees_amount"
    FEES_CURRENCY = "fees_currency"


class PlotCharts(enum.Enum):
    MAIN_CHART = "main-chart"
    SUB_CHART = "sub-chart"


class DisplayedElementTypes(enum.Enum):
    CHART = "chart"
    INPUT = "input"
    TABLE = "table"
    VALUE = "value"
    DICTIONARY = "dictionary"


class DBTables(enum.Enum):
    METADATA = "metadata"
    INPUTS = "inputs"
    PORTFOLIO = "portfolio"
    ORDERS = "all_orders"
    HISTORICAL_ORDERS_UPDATES = "order_updates"
    TRADES = "all_trades"
    TRANSACTIONS = "all_transactions"
    CANDLES = "candles"
    CANDLES_SOURCE = "candles_source"
    CACHE_SOURCE = "cache_source"


class ActivationTopics(enum.Enum):
    """
    Events that can trigger actions
    """

    FULL_CANDLES = "once per bar close"
    IN_CONSTRUCTION_CANDLES = "once per second (Live Price)"
    RECENT_TRADES = "recent trades"
    EVALUATION_CYCLE = "after evaluators"


class TriggerSource(enum.Enum):
    INITIALIZATION = "initialization"
    EVALUATION_MATRIX = "evaluation_matrix"
    EVALUATOR_REFRESH = "evaluator_refresh"
    OHLCV = "ohlcv"
    KLINE = "kline"
    ORDER = "order"
    TRADE = "trade"
    PRICE = "price"
    BALANCE = "balance"
    POSITION = "position"
    CONFIGURATION_UPDATE = "configuration_update"
    MANUAL = "manual"
    UNDEFINED = "undefined"


class DataBaseOrderBy(enum.Enum):
    """
    Database orders
    """

    ASC = "ASC"
    DESC = "DESC"


class DataBaseOperations(enum.Enum):
    """
    Database operators
    """

    SUP = ">"
    INF = "<"
    EQUALS = "="
    INF_EQUALS = f"{INF}{EQUALS}"
    SUP_EQUALS = f"{SUP}{EQUALS}"


class RunDatabases(enum.Enum):
    """
    Database identifiers
    """

    HISTORY = "history"
    LIVE = "live"
    BACKTESTING = "backtesting"
    OPTIMIZER = "optimizer"
    OPTIMIZER_RUNS_SCHEDULE_DB = "runs_schedule"
    RUN_DATA_DB = "run_data"
    PORTFOLIO_VALUE_DB = "portfolio_value"
    HISTORICAL_PORTFOLIO_VALUE = "historical_portfolio_value"
    ORDERS_DB = "orders"
    TRADES_DB = "trades"
    TRANSACTIONS_DB = "transactions"
    EXCHANGES = "exchanges"
    METADATA = "metadata"


class LogicalOperators(enum.Enum):
    """
    Logical operators
    """

    LOWER_THAN = "lower_than"
    HIGHER_THAN = "higher_than"
    LOWER_OR_EQUAL_TO = "lower_or_equal_to"
    HIGHER_OR_EQUAL_TO = "higher_or_equal_to"
    EQUAL_TO = "equal_to"
    DIFFERENT_FROM = "different_from"


class CommunityFeedAttrs(enum.Enum):
    ID = "u"
    STREAM_ID = "i"
    VALUE = "s"
    VERSION = "v"
    TIMESTAMP = "d"
    CHANNEL_TYPE = "t"


class CommunityChannelTypes(enum.Enum):
    CONFIGURATION = "cfg"
    SIGNAL = "t"
    TRADINGVIEW = "tv"
    ALERT = "alert"


class SignalBundlesAttrs(enum.Enum):
    IDENTIFIER = "identifier"
    SIGNALS = "signals"
    VERSION = "version"


class SignalsAttrs(enum.Enum):
    TOPIC = "topic"
    CONTENT = "content"
    DEPENDENCIES = "dependencies"


class SignalDependenciesAttrs(enum.Enum):
    DEPENDENCY = "dependency"


class InitializationEventExchangeTopics(enum.Enum):
    POSITIONS = "positions"
    BALANCE = "balance"
    PROFITABILITY = "profitability"
    ORDERS = "orders"
    TRADES = "trades"
    CONTRACTS = "contracts"
    CANDLES = "candles"
    PRICE = "price"
    ORDER_BOOK = "order_book"
    FUNDING = "funding"
    MARKETS = "markets"


class UserInputTentacleTypes(enum.Enum):
    TRADING_MODE = "trading_mode"
    EVALUATOR = "evaluator"
    EXCHANGE = "exchange"
    BLOCKCHAIN_WALLET = "blockchain_wallet"
    WEB_PLUGIN = "web_plugin"
    AUTOMATION = "automation"
    UNDEFINED = "undefined"


class UserInputTypes(enum.Enum):
    INT = "int"
    FLOAT = "float"
    BOOLEAN = "boolean"
    OPTIONS = "options"
    MULTIPLE_OPTIONS = "multiple-options"
    TEXT = "text"
    OBJECT = "object"
    OBJECT_ARRAY = "object_array"
    STRING_ARRAY = "string_array"


class UserInputEditorOptionsTypes(enum.Enum):
    # source for the available options:
    # https://github.com/json-editor/json-editor#editor-options

    # If set to true, the editor will start collapsed (works for objects and arrays)
    COLLAPSED = "collapsed"
    # If set to true, the "add row" button will be hidden (works for arrays)
    DISABLE_ARRAY_ADD = "disable_array_add"
    # If set to true, all of the "delete" buttons will be hidden (works for arrays)
    DISABLE_ARRAY_DELETE = "disable_array_delete"
    # If set to true, just the "delete all rows"
    #   button will be hidden (works for arrays)
    DISABLE_ARRAY_DELETE_ALL_ROWS = "disable_array_delete_all_rows"
    # If set to true, just the "delete last row"
    #   buttons will be hidden (works for arrays)
    DISABLE_ARRAY_DELETE_LAST_ROW = "disable_array_delete_last_row"
    # If set to true, the "move up/down" buttons will be hidden (works for arrays)
    DISABLE_ARRAY_REORDER = "disable_array_reorder"
    # If set to true, the collapse button will be hidden (works for objects and arrays)
    DISABLE_COLLAPSE = "disable_collapse"
    # If set to true, the Edit JSON button will be hidden (works for objects)
    DISABLE_EDIT_JSON = "disable_edit_json"
    # If set to true, the Edit Properties button will be hidden (works for objects)
    DISABLE_PROPERTIES = "disable_properties"
    # If set to true, array controls (add, delete etc) will be
    #   displayed at top of list (works for arrays)
    ARRAY_CONTROLS_TOP = "array_controls_top"
    # See Enum options (https://github.com/json-editor/json-editor#enum-options)
    ENUM = "enum"
    # An array of display values to use for select box options in the same
    #   order as defined with the enum keyword. Works with schema using enum values.
    ENUM_TITLES = "enum_titles"
    # If set to true, the input will auto expand/contract to fit the content.
    #   Works best with textareas.
    EXPAND_HEIGHT = "expand_height"
    # Explicitly set the number of grid columns (1-12) for the editor
    #   if it's within an object using a grid layout.
    GRID_COLUMNS = "grid_columns"
    # If set to true, the editor will not appear in the UI (works for all types)
    HIDDEN = "hidden"
    # Explicitly set the height of the input element. Should be a valid CSS
    #    width string (e.g. "100px"). Works best with textareas.
    INPUT_HEIGHT = "input_height"
    # Explicitly set the width of the input element. Should be a valid CSS
    #    width string (e.g. "100px"). Works for string, number, and integer data types.
    INPUT_WIDTH = "input_width"
    # If set to true for an object, empty object properties
    #    (i.e. those with falsy values) will not be returned by getValue().
    REMOVE_EMPTY_PROPERTIES = "remove_empty_properties"


class UserInputOtherSchemaValuesTypes(enum.Enum):
    DESCRIPTION = "description"
    DEPENDENCIES = "dependencies"


class ProfileComplexity(enum.Enum):
    EASY = 1
    MEDIUM = 2
    DIFFICULT = 3


class ProfileRisk(enum.Enum):
    LOW = 1
    MODERATE = 2
    HIGH = 3


class ProfileType(enum.Enum):
    LIVE = "live"
    BACKTESTING = "backtesting"


class SignalHistoryTypes(enum.Enum):
    GPT = "gpt"



# pylint: disable=R0913
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.constants as constants

UNSET_EVAL_TYPE = "unset_eval_type_param"


def check_valid_eval_note(
    eval_note,
    eval_type=UNSET_EVAL_TYPE,
    expected_eval_type=None,
    eval_time=None,
    expiry_delay=None,
    current_time=None,
):
    """
    Will also test evaluation type if if eval_type is provided.
    :param eval_note: The evaluation value
    :param eval_type:  The evaluation type
    :param expected_eval_type: The expected type. Default is EVALUATOR_EVAL_DEFAULT_TYPE
    :param eval_time:  The evaluation time
    :param expiry_delay: The allowed evaluation delay
    :param current_time: The current time
    :return: True when evaluation value is valid
    """
    if eval_type != UNSET_EVAL_TYPE and (
        eval_type != expected_eval_type or expected_eval_type is None
    ):
        return False
    return (
        eval_note is not None
        and eval_note is not constants.START_PENDING_EVAL_NOTE
        and (eval_time is None or eval_time + expiry_delay - current_time > 0)
    )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.*
import copy

import pytest

import octobot_commons.profiles as profiles
import octobot_commons.profiles.profile_data as profile_data_import
import octobot_commons.constants as constants
import octobot_commons.enums as enums

from tests.profiles import get_profile_path, profile


@pytest.fixture
def profile_data_dict():
    return {
        'profile_details': {
            'name': 'profile_name 42',
            'id': 'default',
            'version': "42.42.1b",
            'bot_id': "1234-1224-0000",
            'user_id': "plop-id",
            'nested_strategy_config_id': "123-1221"
        },
        'crypto_currencies': [
            {
                'trading_pairs': ['BTC/USDT'],
                'name': 'Bitcoin',
                'enabled': True
            },
            {
                'trading_pairs': ['ETH/USDT'],
                'name': 'ETH',
                'enabled': False
            }
        ], 'exchanges': [
            {
                'exchange_credential_id': '123-plop',
                'exchange_id': '123-exchange',
                'exchange_account_id': '123-exchange_account_id',
                'exchange_type': 'spot',
                'internal_name': 'cryptocom',
                'sandboxed': True,
            }
        ], 'future_exchange_data': {
            'default_leverage': 10,
            'symbol_data': [
                {
                    'symbol': 'BTC/USDT',
                    'leverage': None,
                },
                {
                    'symbol': 'ETH/USDT',
                    'leverage': 5,
                }
            ]
        }, 'trader': {
            'enabled': True
        }, 'trader_simulator': {
            'enabled': False,
            'starting_portfolio': {
                'BTC': 10,
                'USDT': 1000
            },
            'maker_fees': 0.1,
            'taker_fees': 0.1
        }, 'trading': {
            'reference_market': 'BTC',
            'paused': False,
            'minimal_funds': [
                {
                    "asset": "BTC",
                    "available": 12,
                    "total": 12,
                },
                {
                    "asset": "PLOP",
                    "available": 0.1111,
                    "total": 0.2222,
                }
            ],
            'sub_portfolio': {
                "BTC": 0.1,
                "USDT": 1000
            },
            'risk': 0.5,
            'sellable_assets': ["BTC", "ETH"]
        }, 'tentacles': [
            {
                'name': 'plopEvaluator',
                'config': {},
            },
            {
                'name': 'plopEvaluator',
                'config': {
                    'a': True,
                    'other': {
                        'l': [1, 2],
                        'n': None,
                    }
                },
            },
        ], 'options': {
            'values': {
                'plop_key': 'hola senior'
            }
        }, 'backtesting_context': {
            'start_time_delta': 11313.22,
            'update_interval': 444,
            'starting_portfolio': {
                'plop_key': 'hola senior'
            },
            'exchanges': [
                'binance', 'kucoin'
            ]
        }
    }


@pytest.fixture
def min_profile_data_dict():
    return {
        'profile_details': {
            'name': 'min_profile',
        },
        'crypto_currencies': [
            {
                'trading_pairs': ['BTC/USDT'],
            },
            {
                'trading_pairs': ['ETH/USDT'],
                'enabled': False
            }
        ], 'trading': {
            'reference_market': 'BTC',
        }, 'tentacles': [
            {
                'name': 'plopEvaluator',
                'config': {},
            },
            {
                'name': 'plopEvaluator',
                'config': {
                    'a': True,
                    'other': {
                        'l': [1, 2],
                        'n': None,
                    }
                },
            },
        ],
        'backtesting_context': {
            'start_time_delta': 11313.22
        }, 'options': {
            'values': {
                'plop_key': 'hola !!!',
                'jour': 'nuit',
            }
        }
    }


def test_from_profile(profile):
    profile_data = profiles.ProfileData.from_profile(profile.read_config())
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "default"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.exchanges == []
    assert profile_data.trader.enabled is False
    assert profile_data.trader_simulator.enabled is True
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.tentacles == []


def test_to_profile(profile):
    profile_data = profiles.ProfileData.from_profile(profile.read_config())
    created_profile = profile_data.to_profile("plop_path")
    # force missing values
    for crypto_data in profile.config[constants.CONFIG_CRYPTO_CURRENCIES].values():
        crypto_data[constants.CONFIG_ENABLED_OPTION] = crypto_data.get(constants.CONFIG_ENABLED_OPTION, True)
    # remove not stored values
    profile.config[constants.CONFIG_EXCHANGES] = {}
    profile.avatar = profile.description = ""
    profile.complexity = enums.ProfileComplexity.MEDIUM
    profile.risk = enums.ProfileRisk.MODERATE
    profile.profile_type = enums.ProfileType.LIVE
    profile.origin_url = None
    # if both parsing and transforming return the same profile as original one, the whole chain works
    profile_dict = profile.as_dict()
    assert profile_dict == created_profile.as_dict()


def test_from_dict(profile_data_dict):
    # use second MinimalFund syntax
    profile_data_dict = copy.deepcopy(profile_data_dict)
    profile_data_dict["trading"]['minimal_funds'].append(
        {
            "asset": "ETH",
            "value": 111.2,
        }
    )
    profile_data = profiles.ProfileData.from_dict(profile_data_dict)
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "profile_name 42"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.future_exchange_data.default_leverage == 10
    assert profile_data.future_exchange_data.symbol_data[0].symbol == "BTC/USDT"
    assert profile_data.future_exchange_data.symbol_data[0].leverage == None
    assert profile_data.future_exchange_data.symbol_data[1].symbol == "ETH/USDT"
    assert profile_data.future_exchange_data.symbol_data[1].leverage == 5
    assert profile_data.exchanges[0].exchange_credential_id == "123-plop"
    assert profile_data.exchanges[0].exchange_account_id == "123-exchange_account_id"
    assert profile_data.exchanges[0].internal_name == "cryptocom"
    assert profile_data.exchanges[0].sandboxed is True
    assert profile_data.exchanges[0].exchange_type == "spot"
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.trading.minimal_funds == [
        profiles.MinimalFund("BTC", 12, 12),
        profiles.MinimalFund("PLOP", 0.1111, 0.2222),
        profiles.MinimalFund("ETH", 111.2, 111.2),
    ]
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]


def test_from_min_dict(min_profile_data_dict):
    profile_data = profiles.ProfileData.from_dict(min_profile_data_dict)
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.name == "min_profile"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.crypto_currencies[0].name is None
    assert profile_data.crypto_currencies[0].enabled is True
    assert profile_data.crypto_currencies[1].trading_pairs == ['ETH/USDT']
    assert profile_data.crypto_currencies[1].name is None
    assert profile_data.crypto_currencies[1].enabled is False
    assert profile_data.future_exchange_data.default_leverage is None
    assert profile_data.future_exchange_data.symbol_data == []
    assert profile_data.exchanges == []
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {}
    assert profile_data.trading.risk == 1
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]
    assert profile_data.options.values["jour"] == "nuit"
    assert profile_data.options.values["plop_key"] == "hola !!!"
    full_profile_data_dict = profile_data.to_dict(include_default_values=True)
    assert len(full_profile_data_dict) > len(min_profile_data_dict)
    profile_data_dict = profile_data.to_dict(include_default_values=False)
    # default values in values but: keys are present except for exchanges, which content is empty (default)
    full_profile_data_dict_keys_without_exchange = list(full_profile_data_dict.keys())
    full_profile_data_dict_keys_without_exchange.remove("exchanges")
    assert sorted(list(profile_data_dict.keys()) + ["_updated_fields", ]) == sorted(full_profile_data_dict_keys_without_exchange)


def test_from_dict_objects(profile_data_dict):
    profile_data_objects = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data = profiles.ProfileData.from_dict({
        "profile_details": profile_data_objects.profile_details,
        "crypto_currencies": profile_data_objects.crypto_currencies,
        "exchanges": profile_data_objects.exchanges,
        "trader": profile_data_objects.trader,
        "trader_simulator": profile_data_objects.trader_simulator,
        "trading": profile_data_objects.trading,
        "tentacles": profile_data_objects.tentacles,
        "options": profile_data_objects.options,
    })
    # check one element per attribute to be sure it's all parsed
    assert profile_data.profile_details.version == "42.42.1b"
    assert profile_data.profile_details.nested_strategy_config_id == "123-1221"
    assert profile_data.crypto_currencies[0].trading_pairs == ['BTC/USDT']
    assert profile_data.exchanges[0].exchange_credential_id == "123-plop"
    assert profile_data.exchanges[0].exchange_account_id == "123-exchange_account_id"
    assert profile_data.trader.enabled is True
    assert profile_data.trader_simulator.enabled is False
    assert profile_data.trader_simulator.starting_portfolio == {'BTC': 10, 'USDT': 1000}
    assert profile_data.trading.risk == 0.5
    assert profile_data.trading.paused is False
    assert profile_data.tentacles[0].name == "plopEvaluator"
    assert profile_data.tentacles[1].config["other"]["l"] == [1, 2]
    assert profile_data.options.values['plop_key'] == 'hola senior'
    assert profile_data.backtesting_context == profile_data_import.BacktestingContext()


def test_to_dict(profile_data_dict):
    profile_data = profiles.ProfileData.from_dict(profile_data_dict)
    # if both parsing and transforming return the same profile as original one, the whole chain works
    assert profile_data_dict == _remove_updated_fields(profile_data.to_dict())

    dict_without_default_values = profile_data.to_dict(include_default_values=False)
    assert profile_data_dict != _remove_updated_fields(dict_without_default_values)

    # ensure no empty elements
    assert len(dict_without_default_values) == len(profile_data_dict)
    for values in dict_without_default_values.values():
        assert len(values)


def test_get_update(profile_data_dict):
    """Test get_update() method with various field changes and edge cases"""
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    
    # Test simple field changes
    profile_data_2.profile_details.name = "new_profile_name"
    profile_data_2.trading.risk = 0.8
    update = profile_data_1.get_update(profile_data_2)
    assert update.profile_details.name == "new_profile_name"
    assert update.profile_details._updated_fields == ["name"]
    assert update.trading.risk == 0.8
    assert update.trading._updated_fields == ["risk"]
    assert "profile_details" in update._updated_fields
    assert "trading" in update._updated_fields
    
    # Test nested object changes
    profile_data_2.profile_details.version = "1.0.0"
    profile_data_2.trading.reference_market = "USDT"
    update = profile_data_1.get_update(profile_data_2)
    assert set(update.profile_details._updated_fields) == {"name", "version"}
    assert set(update.trading._updated_fields) == {"risk", "reference_market"}
    
    # Test list field changes
    profile_data_2.crypto_currencies[0].enabled = False
    profile_data_2.crypto_currencies[0].trading_pairs = ["BTC/USDT", "ETH/USDT"]
    new_exchange = profile_data_import.ExchangeData(
        exchange_type="futures",
        internal_name="binance",
        sandboxed=False
    )
    profile_data_2.exchanges.append(new_exchange)
    profile_data_2.tentacles[1].config["new_key"] = "new_value"
    update = profile_data_1.get_update(profile_data_2)
    assert update.crypto_currencies[0].enabled is False
    assert update.crypto_currencies[0]._updated_fields == ["trading_pairs", "enabled"]
    assert len(update.exchanges) == len(profile_data_2.exchanges)
    assert update.exchanges[-1].internal_name == "binance"
    assert "crypto_currencies" in update._updated_fields
    assert "exchanges" in update._updated_fields
    assert "tentacles" in update._updated_fields
    
    # Test no changes
    profile_data_3 = copy.deepcopy(profile_data_2)
    update = profile_data_2.get_update(profile_data_3)
    assert update._updated_fields == []
    
    # Test empty to populated
    profile_data_empty = profiles.ProfileData()
    update = profile_data_empty.get_update(profile_data_1)
    assert "profile_details" in update._updated_fields
    assert "crypto_currencies" in update._updated_fields
    assert "trading" in update._updated_fields


def test_update(profile_data_dict):
    """Test update() method with various field changes, lists, and edge cases"""
    # Test simple field updates
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.profile_details.name = "updated_name"
    profile_data_2.trading.risk = 0.9
    update = profile_data_1.get_update(profile_data_2)
    original_name = profile_data_1.profile_details.name
    original_risk = profile_data_1.trading.risk
    profile_data_1.update(update)
    assert profile_data_1.profile_details.name == "updated_name"
    assert profile_data_1.trading.risk == 0.9
    assert profile_data_1.profile_details.name != original_name
    assert profile_data_1.trading.risk != original_risk
    
    # Test nested object updates
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.profile_details.version = "2.0.0"
    profile_data_2.trader.enabled = False
    profile_data_2.trader_simulator.enabled = True
    profile_data_2.trader_simulator.maker_fees = 0.2
    update = profile_data_1.get_update(profile_data_2)
    profile_data_1.update(update)
    assert profile_data_1.profile_details.version == "2.0.0"
    assert profile_data_1.trader.enabled is False
    assert profile_data_1.trader_simulator.enabled is True
    assert profile_data_1.trader_simulator.maker_fees == 0.2
    
    # Test list element modifications
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.crypto_currencies[0].enabled = False
    profile_data_2.crypto_currencies[0].trading_pairs.append("LTC/USDT")
    profile_data_2.exchanges[0].sandboxed = False
    update = profile_data_1.get_update(profile_data_2)
    original_enabled = profile_data_1.crypto_currencies[0].enabled
    profile_data_1.update(update)
    assert profile_data_1.crypto_currencies[0].enabled is False
    assert profile_data_1.crypto_currencies[0].enabled != original_enabled
    assert "LTC/USDT" in profile_data_1.crypto_currencies[0].trading_pairs
    assert profile_data_1.exchanges[0].sandboxed is False
    
    # Test adding list elements
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    new_crypto = profile_data_import.CryptoCurrencyData(
        trading_pairs=["LTC/USDT"],
        name="Litecoin",
        enabled=True
    )
    profile_data_2.crypto_currencies.append(new_crypto)
    new_tentacle = profile_data_import.TentaclesData(
        name="NewEvaluator",
        config={"key": "value"}
    )
    profile_data_2.tentacles.append(new_tentacle)
    update = profile_data_1.get_update(profile_data_2)
    original_crypto_count = len(profile_data_1.crypto_currencies)
    original_tentacle_count = len(profile_data_1.tentacles)
    profile_data_1.update(update)
    assert len(profile_data_1.crypto_currencies) == original_crypto_count + 1
    assert profile_data_1.crypto_currencies[-1].name == "Litecoin"
    assert len(profile_data_1.tentacles) == original_tentacle_count + 1
    assert profile_data_1.tentacles[-1].name == "NewEvaluator"
    
    # Test removing list elements
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.crypto_currencies.pop()
    profile_data_2.tentacles.pop()
    update = profile_data_1.get_update(profile_data_2)
    original_crypto_count = len(profile_data_1.crypto_currencies)
    original_tentacle_count = len(profile_data_1.tentacles)
    profile_data_1.update(update)
    assert len(profile_data_1.crypto_currencies) == original_crypto_count - 1
    assert len(profile_data_1.tentacles) == original_tentacle_count - 1
    
    # Test future exchange data updates
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.future_exchange_data.default_leverage = 20
    profile_data_2.future_exchange_data.symbol_data[0].leverage = 15
    update = profile_data_1.get_update(profile_data_2)
    original_leverage = profile_data_1.future_exchange_data.default_leverage
    profile_data_1.update(update)
    assert profile_data_1.future_exchange_data.default_leverage == 20
    assert profile_data_1.future_exchange_data.default_leverage != original_leverage
    assert profile_data_1.future_exchange_data.symbol_data[0].leverage == 15
    
    # Test trading minimal_funds updates
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    new_fund = profile_data_import.MinimalFund("ETH", 50.0, 50.0)
    profile_data_2.trading.minimal_funds.append(new_fund)
    profile_data_2.trading.minimal_funds[0].available = 20.0
    update = profile_data_1.get_update(profile_data_2)
    original_fund_count = len(profile_data_1.trading.minimal_funds)
    profile_data_1.update(update)
    assert len(profile_data_1.trading.minimal_funds) == original_fund_count + 1
    assert profile_data_1.trading.minimal_funds[0].available == 20.0
    assert profile_data_1.trading.minimal_funds[-1].asset == "ETH"
    
    # Test update without _updated_fields (should not update)
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    profile_data_2.profile_details.name = "should_not_update"
    profile_data_2.trading.risk = 0.99
    update = profile_data_1.get_update(profile_data_2)
    update._updated_fields = []
    update.profile_details._updated_fields = []
    update.trading._updated_fields = []
    original_name = profile_data_1.profile_details.name
    original_risk = profile_data_1.trading.risk
    profile_data_1.update(update)
    assert profile_data_1.profile_details.name == original_name
    assert profile_data_1.trading.risk == original_risk


def test_get_update_and_update_roundtrip(profile_data_dict):
    """Test that get_update() and update() work together correctly"""
    profile_data_1 = profiles.ProfileData.from_dict(profile_data_dict)
    profile_data_2 = copy.deepcopy(profile_data_1)
    
    # Make various changes
    profile_data_2.profile_details.name = "roundtrip_name"
    profile_data_2.profile_details.version = "3.0.0"
    profile_data_2.trading.risk = 0.7
    profile_data_2.trading.reference_market = "ETH"
    profile_data_2.crypto_currencies[0].enabled = False
    profile_data_2.trader.enabled = True
    
    # Get update and apply it
    update = profile_data_1.get_update(profile_data_2)
    profile_data_1.update(update)
    
    # Verify profile_data_1 matches profile_data_2 for changed fields
    assert profile_data_1.profile_details.name == profile_data_2.profile_details.name
    assert profile_data_1.profile_details.version == profile_data_2.profile_details.version
    assert profile_data_1.trading.risk == profile_data_2.trading.risk
    assert profile_data_1.trading.reference_market == profile_data_2.trading.reference_market
    assert profile_data_1.crypto_currencies[0].enabled == profile_data_2.crypto_currencies[0].enabled
    assert profile_data_1.trader.enabled == profile_data_2.trader.enabled


def _remove_updated_fields(profile_data_dict: dict) -> dict:
    profile_data_dict.pop("_updated_fields", None)
    for value in profile_data_dict.values():
        if isinstance(value, dict):
            _remove_updated_fields(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _remove_updated_fields(item)
    return profile_data_dict



# pylint: disable=R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import secrets
from typing import Optional, Tuple

from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.backends import default_backend


def generate_rsa_key_pair(key_size: int = 4096) -> Tuple[bytes, bytes]:
    """Generate an RSA key pair

    This function generates a private and public RSA key pair equivalent to:
    - openssl genrsa -out user_private_key.pem 4096
    - openssl rsa -in user_private_key.pem -pubout -out user_public_key.pem

    :param key_size: The size of the RSA key in bits. Defaults to 4096.
    :type key_size: int
    :return: A tuple containing (private_key_pem, public_key_pem) as bytes.
    :rtype: Tuple[bytes, bytes]
    :raises ValueError: If key_size is less than 2048 bits (security best practice).
    """
    if key_size < 2048:
        raise ValueError("RSA key size must be at least 2048 bits for security.")

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
    )

    # Serialize private key to PEM format
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    # Extract and serialize public key to PEM format
    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return private_key_pem, public_key_pem


def generate_aes_key(key_size: int = 32) -> bytes:
    """Generate a random AES key

    This function generates random bytes equivalent to:
    - openssl rand -out aes_key_s1.bin 32

    :param key_size: The size of the AES key in bytes. Defaults to 32 (256 bits).
    :type key_size: int
    :return: Random bytes suitable for use as an AES key.
    :rtype: bytes
    :raises ValueError: If key_size is less than 1.
    """
    if key_size < 1:
        raise ValueError("Key size must be at least 1 byte.")

    return secrets.token_bytes(key_size)


def generate_iv(iv_size: int = 12) -> bytes:
    """Generate a random initialization vector (IV)

    This function generates random bytes equivalent to:
    - openssl rand -out iv_s1.bin 12

    :param iv_size: The size of the IV in bytes. Defaults to 12 (standard for AES-GCM).
    :type iv_size: int
    :return: Random bytes suitable for use as an IV.
    :rtype: bytes
    :raises ValueError: If iv_size is less than 1.
    """
    if iv_size < 1:
        raise ValueError("IV size must be at least 1 byte.")

    return secrets.token_bytes(iv_size)


def rsa_decrypt_aes_key(encrypted_aes_key: bytes, private_key_pem: bytes) -> bytes:
    """Decrypt an RSA-encrypted AES key

    This function decrypts an AES key equivalent to:
    - openssl rsautl -decrypt -in aes_key.enc -oaep -inkey private_key.pem -out aes_key.bin

    :param encrypted_aes_key: The RSA-encrypted AES key.
    :type encrypted_aes_key: bytes
    :param private_key_pem: The RSA private key in PEM format.
    :type private_key_pem: bytes
    :return: The decrypted AES key as bytes.
    :rtype: bytes
    :raises ValueError: If the private key cannot be loaded or is invalid.
    """
    # Load the private key from PEM format
    private_key = serialization.load_pem_private_key(
        private_key_pem,
        password=None,
    )

    if not isinstance(private_key, rsa.RSAPrivateKey):
        raise ValueError("Private key must be an RSA key.")

    # Decrypt using OAEP padding
    decrypted_key = private_key.decrypt(
        encrypted_aes_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    return decrypted_key


def rsa_encrypt_aes_key(aes_key: bytes, public_key_pem: bytes) -> bytes:
    """Encrypt an AES key using RSA public key

    This function encrypts an AES key equivalent to:
    - openssl rsautl -encrypt -in aes_key.bin -oaep -pubin -inkey public_key.pem -out aes_key.enc

    :param aes_key: The AES key to encrypt.
    :type aes_key: bytes
    :param public_key_pem: The RSA public key in PEM format.
    :type public_key_pem: bytes
    :return: The encrypted AES key as bytes.
    :rtype: bytes
    :raises ValueError: If the public key cannot be loaded or is invalid.
    """
    # Load the public key from PEM format
    public_key = serialization.load_pem_public_key(public_key_pem)

    if not isinstance(public_key, rsa.RSAPublicKey):
        raise ValueError("Public key must be an RSA key.")

    # Encrypt using OAEP padding
    encrypted_key = public_key.encrypt(
        aes_key,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )

    return encrypted_key


def aes_gcm_decrypt(
    encrypted_data: bytes,
    aes_key: bytes,
    iv: bytes,
    associated_data: Optional[bytes] = None,
) -> bytes:
    """Decrypt data using AES-256-GCM

    This function decrypts data equivalent to:
    - openssl enc -aes-256-gcm -d -in data.enc -out data.bin -K $AES_KEY -iv $IV

    :param encrypted_data: The encrypted data (ciphertext + authentication tag).
    :type encrypted_data: bytes
    :param aes_key: The AES key (32 bytes for AES-256).
    :type aes_key: bytes
    :param iv: The initialization vector (12 bytes for AES-GCM).
    :type iv: bytes
    :param associated_data: Optional associated data (AAD) for authenticated encryption.
    :type associated_data: Optional[bytes]
    :return: The decrypted plaintext data.
    :rtype: bytes
    :raises ValueError: If key or IV sizes are invalid.
    """
    if len(aes_key) != 32:
        raise ValueError("AES key must be 32 bytes (256 bits) for AES-256-GCM.")

    if len(iv) != 12:
        raise ValueError("IV must be 12 bytes for AES-GCM.")

    # AESGCM expects the authentication tag to be appended to the ciphertext
    # OpenSSL GCM format: ciphertext + 16-byte tag
    if len(encrypted_data) < 16:
        raise ValueError(
            "Encrypted data too short (must include 16-byte authentication tag)."
        )

    # Create AESGCM cipher
    aesgcm = AESGCM(aes_key)

    # Decrypt with authentication (encrypted_data should be ciphertext + 16-byte tag)
    plaintext = aesgcm.decrypt(iv, encrypted_data, associated_data)

    return plaintext


def aes_gcm_encrypt(
    plaintext: bytes, aes_key: bytes, iv: bytes, associated_data: Optional[bytes] = None
) -> bytes:
    """Encrypt data using AES-256-GCM

    This function encrypts data equivalent to:
    - openssl enc -aes-256-gcm -in data.bin -out data.enc -K $AES_KEY -iv $IV

    :param plaintext: The plaintext data to encrypt.
    :type plaintext: bytes
    :param aes_key: The AES key (32 bytes for AES-256).
    :type aes_key: bytes
    :param iv: The initialization vector (12 bytes for AES-GCM).
    :type iv: bytes
    :param associated_data: Optional associated data (AAD) for authenticated encryption.
    :type associated_data: Optional[bytes]
    :return: The encrypted data (ciphertext + 16-byte authentication tag).
    :rtype: bytes
    :raises ValueError: If key or IV sizes are invalid.
    """
    if len(aes_key) != 32:
        raise ValueError("AES key must be 32 bytes (256 bits) for AES-256-GCM.")

    if len(iv) != 12:
        raise ValueError("IV must be 12 bytes for AES-GCM.")

    # Create AESGCM cipher
    aesgcm = AESGCM(aes_key)

    # Encrypt with authentication (returns ciphertext + tag)
    ciphertext_with_tag = aesgcm.encrypt(iv, plaintext, associated_data)

    return ciphertext_with_tag


def pbkdf2_derive_key_from_pin(
    pin: str, salt: bytes, iterations: int = 200000, key_size: int = 32
) -> bytes:
    """Derive a cryptographic key from a PIN using PBKDF2

    This function derives a key from a user PIN using PBKDF2-HMAC-SHA256.
    High iteration count is recommended for PINs due to their low entropy.

    :param pin: The user PIN (4-6 digits recommended).
    :type pin: str
    :param salt: Random salt bytes (should be unique per encryption, 16 bytes recommended).
    :type salt: bytes
    :param iterations: Number of PBKDF2 iterations. Defaults to 200000 (high for PIN security).
    :type iterations: int
    :param key_size: Size of the derived key in bytes. Defaults to 32 (256 bits for AES-256).
    :type key_size: int
    :return: The derived key as bytes.
    :rtype: bytes
    :raises ValueError: If salt is empty or iterations/key_size are invalid.
    """
    if len(salt) < 1:
        raise ValueError("Salt must be at least 1 byte.")
    if iterations < 1:
        raise ValueError("Iterations must be at least 1.")
    if key_size < 1:
        raise ValueError("Key size must be at least 1 byte.")

    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=key_size,
        salt=salt,
        iterations=iterations,
        backend=default_backend(),
    )
    return kdf.derive(pin.encode("utf-8"))


def pbkdf2_encrypt_aes_key(
    aes_key: bytes,
    pin: str,
    salt: Optional[bytes] = None,
    iterations: int = 200000,
) -> Tuple[bytes, bytes, bytes]:
    """Encrypt an AES key using a PIN-derived key via PBKDF2

    This function encrypts an AES key using a key derived from a user PIN.
    The salt and IV are generated randomly and must be stored for decryption.

    :param aes_key: The AES key to encrypt (32 bytes for AES-256).
    :type aes_key: bytes
    :param pin: The user PIN (4-6 digits recommended).
    :type pin: str
    :param salt: Optional salt bytes. If None, a random 16-byte salt is generated.
    :type salt: Optional[bytes]
    :param iterations: Number of PBKDF2 iterations. Defaults to 200000.
    :type iterations: int
    :return: A tuple containing (encrypted_key, salt, iv).
    :rtype: Tuple[bytes, bytes, bytes]
    :raises ValueError: If AES key size is invalid.
    """
    if len(aes_key) != 32:
        raise ValueError("AES key must be 32 bytes (256 bits) for AES-256.")

    if salt is None:
        salt = secrets.token_bytes(16)

    # Derive key from PIN
    derived_key = pbkdf2_derive_key_from_pin(pin, salt, iterations, key_size=32)

    # Generate IV for AES-GCM
    iv = generate_iv()

    # Encrypt AES key using derived key
    encrypted_key = aes_gcm_encrypt(aes_key, derived_key, iv)

    return encrypted_key, salt, iv


def pbkdf2_decrypt_aes_key(
    encrypted_aes_key: bytes, pin: str, salt: bytes, iv: bytes, iterations: int = 200000
) -> bytes:
    """Decrypt an AES key using a PIN-derived key via PBKDF2

    This function decrypts an AES key that was encrypted with pbkdf2_encrypt_aes_key.

    :param encrypted_aes_key: The encrypted AES key.
    :type encrypted_aes_key: bytes
    :param pin: The user PIN used for encryption.
    :type pin: str
    :param salt: The salt used during encryption.
    :type salt: bytes
    :param iv: The IV used during encryption.
    :type iv: bytes
    :param iterations: Number of PBKDF2 iterations (must match encryption). Defaults to 200000.
    :type iterations: int
    :return: The decrypted AES key as bytes.
    :rtype: bytes
    :raises ValueError: If decryption fails (wrong PIN, corrupted data, etc.).
    """
    # Derive key from PIN (same parameters as encryption)
    derived_key = pbkdf2_derive_key_from_pin(pin, salt, iterations, key_size=32)

    # Decrypt AES key using derived key
    try:
        decrypted_key = aes_gcm_decrypt(encrypted_aes_key, derived_key, iv)
    except Exception as e:
        raise ValueError(
            "Failed to decrypt AES key. Wrong PIN or corrupted data."
        ) from e

    return decrypted_key



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from datetime import datetime, timezone


LOCAL_TIMEZONE = datetime.now().astimezone().tzinfo


def convert_timestamp_to_datetime(
    timestamp: float, time_format: str = "%d/%m/%y %H:%M", local_timezone: bool = False
) -> str:
    """
    Convert a timestamp to a human readable string
    :param timestamp: the timestamp to convert
    :param time_format: the time format
    :param local_timezone: if the local timezone should be used
    :return: the created readable string
    """
    return datetime.fromtimestamp(
        timestamp, tz=(LOCAL_TIMEZONE if local_timezone else timezone.utc)
    ).strftime(time_format)


def convert_timestamps_to_datetime(
    timestamps: list[float],
    time_format: str = "%d/%m/%y %H:%M",
    local_timezone: bool = False,
) -> list[str]:
    """
    Convert multiple timestamps to datetime objects
    :param timestamps: the timestamp to convert list
    :param time_format: the time format
    :param local_timezone: if the local timezone should be used
    :return: the created datetime objects
    """
    return [
        convert_timestamp_to_datetime(
            timestamp, time_format=time_format, local_timezone=local_timezone
        )
        for timestamp in timestamps
    ]


def is_valid_timestamp(timestamp: float) -> bool:
    """
    Check if the timestamp is valid
    :param timestamp: the timestamp to check
    :return: the check result
    """
    if timestamp:
        try:
            datetime.fromtimestamp(timestamp)
        except (OSError, ValueError, OverflowError, TypeError):
            return False
    return True


def get_now_time(
    time_format: str = "%Y-%m-%d %H:%M:%S", local_timezone: bool = True
) -> str:
    """
    Get the current time
    :param time_format: the time format
    :return: the current timestamp
    """
    return datetime.now(
        tz=(LOCAL_TIMEZONE if local_timezone else timezone.utc)
    ).strftime(time_format)


def datetime_to_timestamp(
    date_time_str: str, date_time_format: str, local_timezone: bool = True
) -> float:
    """
    Convert a datetime to timestamp
    :param date_time_str: the datetime string
    :param date_time_format: the datetime format
    :return: the timestamp
    """
    return create_datetime_from_string(
        date_time_str, date_time_format, local_timezone=local_timezone
    ).timestamp()


def create_datetime_from_string(
    date_time_str: str, date_time_format: str, local_timezone: bool = True
) -> datetime:
    """
    Convert a string to datetime
    :param date_time_str: the datetime string
    :param date_time_format: the datetime format
    :return: the converted datetime
    """
    # force local timezone or parsing might fail
    return datetime.strptime(date_time_str, date_time_format).replace(
        tzinfo=LOCAL_TIMEZONE if local_timezone else timezone.utc
    )



# pylint: disable=R1725,W0221
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.tree.base_tree as base_tree
import octobot_commons.logging as logging


class EventTreeNode(base_tree.BaseTreeNode):
    """
    Node element of an EventTreeNode. self.node_value is an asyncio.Event() that is triggered when all of its children
    are triggered or is manually triggered. Adding an unset child will clear self. Children updates will overwrite
    any manual trigger
    """

    __slots__ = [
        "_parent",
        "_logger",
    ]

    def __init__(self, *_, triggered=False, **__):
        super().__init__(asyncio.Event(), asyncio.Event)
        if triggered:
            self._trigger()
        self._parent = None
        self._logger = logging.get_logger(self.__class__.__name__)

    def bind_parent(self, parent):
        """
        Set the parent node and propagate the local state to the parent node
        """
        self._parent = parent
        self._propagate()

    def is_triggered(self):
        """
        Return True if the local event is set
        """
        return self.node_value.is_set()

    async def wait(self):
        """
        Wait till the local event is triggered
        """
        await self.node_value.wait()

    def trigger(self):
        """
        Trigger the local event, propagate to the parent if any change
        """
        if not self.is_triggered():
            self._trigger_and_log()
            self._propagate()

    def _trigger_and_log(self):
        """
        Set the event and log
        """
        self._trigger()
        path_to_root = self.get_path_to_root()
        if path_to_root:
            self._logger.debug(f"Event triggered for {'|'.join(path_to_root)}")

    def _trigger(self):
        """
        Set the event and log
        """
        self.node_value.set()

    def _clear(self):
        """
        Clear the event and log
        """
        self.node_value.clear()
        path_to_root = self.get_path_to_root()
        if path_to_root:
            self._logger.debug(f"Event cleared for {'|'.join(path_to_root)}")

    def get_parent(self):
        """
        Return self._parent
        """
        return self._parent

    def get_path_to_root(self):
        """
        Return the path to self from the root event
        """
        node = self
        path = []
        while node.get_parent() is not None:
            parent = node.get_parent()
            try:
                path = [parent.get_child_key(node)] + path
            except KeyError:
                return path
            node = parent
        return path

    def get_child_key(self, child_to_find):
        """
        Return the key of the given child in the current children
        """
        for key, child in self.children.items():
            if child is child_to_find:
                return key
        raise KeyError(child_to_find)

    def clear(self):
        """
        Clear the local event, propagate to the parent if any change
        """
        if self.is_triggered():
            self._clear()
            self._propagate()

    def set_child(self, key, child):
        """
        Set a child at the given key
        """
        super(EventTreeNode, self).set_child(key, child)
        self.on_child_change()

    def pop_child(self, key, default):
        """
        Pop the child the given key
        """
        node = super(EventTreeNode, self).pop_child(key, default)
        self.on_child_change()
        return node

    def _untriggered_children(self):
        """
        Return the list of children events that are not triggered
        """
        return [key for key, child in self.children.items() if not child.is_triggered()]

    def on_child_change(self):
        """
        Trigger or clear the local event depending on children states
        then propagate to the parent if any change
        """
        if not self.children:
            # do not change event when no children
            return
        should_be_triggered = True
        untriggered_children = self._untriggered_children()
        if untriggered_children:
            should_be_triggered = False
            if not self.is_triggered():
                self._logger.debug(
                    f"Waiting children trigger for {'|'.join(self.get_path_to_root())}. "
                    f"Untriggered children: {untriggered_children}"
                )
        if should_be_triggered != self.is_triggered():
            if self.is_triggered():
                self._clear()
            else:
                self._trigger_and_log()
            self._propagate()

    def _propagate(self):
        """
        Calls parent's on_child_change
        """
        if self._parent is not None:
            self._parent.on_child_change()


class EventTree(base_tree.BaseTree):
    """
    Tree based on EventTreeNode where each node's event is synchronized with its children to be triggered when all
    of its children are triggered. Adding an untriggered child will untrigger the parent node.
    """

    TREE_NODE_CLASS = EventTreeNode

    def create_node_at_path(self, path, triggered):
        """
        Set the node attributes
        Creates the node if it doesn't exist
        :param path: the node path (as a list of string)
        :param triggered: if the created node event should be initially triggered
        For example:
        - If you created a first node with the path ["my-parent-node"]
        - You can create a child node of my-parent-node by using ["my-parent-node", "my-new-child-node"] as `path`
        :return: void
        """
        self._set_node(self.get_or_create_node(path, triggered=triggered))

    def child_factory(self, node, key, triggered=False, **kwargs):
        """
        Create a new child an associate it to the given key
        """
        new_child = super(EventTree, self).child_factory(
            node, key, triggered=triggered, **kwargs
        )
        new_child.bind_parent(node)
        return new_child



# pylint: disable=R0913, W0718
# Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import contextlib
import time

import octobot_commons.singleton as singleton
import octobot_commons.logging as logging
import octobot_commons.errors as errors
import octobot_commons.signals.signals_emitter as signals_emitter
import octobot_commons.signals.signal_bundle_builder as signal_bundle_builder
import octobot_commons.signals.signal_builder_wrapper as signal_builder_wrapper


class SignalPublisher(singleton.Singleton):
    DEFAULT_SIGNAL_BUILDER_CLASS = signal_bundle_builder.SignalBundleBuilder

    def __init__(self):
        self._signal_builder_wrappers = {}
        self._timeout_watcher_tasks = {}

    def get_signal_bundle_builder(
        self, wrapper_key: str
    ) -> signal_builder_wrapper.SignalBuilderWrapper:
        """
        Return the SignalBuilderWrapper registered under the given key
        """
        try:
            return self._signal_builder_wrappers[wrapper_key].signal_bundle_builder
        except KeyError as err:
            raise errors.MissingSignalBuilder from err

    @contextlib.asynccontextmanager
    async def remote_signal_bundle_builder(
        self,
        wrapper_key: str,
        identifier: str,
        timeout: float = signal_builder_wrapper.SignalBuilderWrapper.NO_TIMEOUT_VALUE,
        signal_builder_class=DEFAULT_SIGNAL_BUILDER_CLASS,
        builder_args=None,
    ):
        """
        Context manager ensuring that any signal under the given key is buildable and sent
        when context manager is closing.
        Use signal_builder_class to specify signal builders to create
        """
        signal_builder_wrap = None
        try:
            signal_builder_wrap = self._create_or_get_signal_builder_wrapper(
                wrapper_key, identifier, timeout, signal_builder_class, builder_args
            )
            signal_builder_wrap.register_user()
            self._register_timeout_if_any(wrapper_key)
            yield signal_builder_wrap.signal_bundle_builder
            # send the full signal when the last user is done
            if signal_builder_wrap.has_single_user():
                await self._emit_signal_if_necessary(signal_builder_wrap)
        finally:
            if signal_builder_wrap is not None:
                if signal_builder_wrap.has_single_user():
                    self._unregister_timeout(wrapper_key)
                    self._signal_builder_wrappers.pop(wrapper_key, None)
                else:
                    signal_builder_wrap.unregister_user()

    def stop(self):
        """
        Stop all timeout tasks and clear any remaining registered wrapper
        :return:
        """
        logging.get_logger(self.__class__.__name__).debug("Stopping ...")
        for task in self._timeout_watcher_tasks.values():
            task.cancel()
        self._timeout_watcher_tasks = {}
        self._signal_builder_wrappers = {}
        logging.get_logger(self.__class__.__name__).debug("Stopped ...")

    def _create_or_get_signal_builder_wrapper(
        self,
        wrapper_key: str,
        identifier: str,
        timeout: float,
        signal_builder_class,
        builder_args: tuple,
    ) -> signal_builder_wrapper.SignalBuilderWrapper:
        if wrapper_key in self._signal_builder_wrappers:
            return self._signal_builder_wrappers[wrapper_key]
        self._signal_builder_wrappers[wrapper_key] = (
            signal_builder_wrapper.SignalBuilderWrapper(
                identifier, signal_builder_class, timeout, builder_args
            )
        )
        return self._signal_builder_wrappers[wrapper_key]

    async def _emit_signal_if_necessary(self, signal_builder_wrap):
        # check has_single_user in case the same builder is used multiple times at once
        if (
            not signal_builder_wrap.signal_bundle_builder.is_empty()
            and not signal_builder_wrap.is_being_emitted
        ):
            try:
                signal_builder_wrap.is_being_emitted = True
                await signals_emitter.emit_signal_bundle(
                    signal_builder_wrap.signal_bundle_builder.build()
                )
            except Exception as err:
                logging.get_logger(self.__class__.__name__).exception(
                    err, True, f"Unexpected error when emitting signal: {err}"
                )
            finally:
                # always reset builder after emitting to avoid emitting the same signal twice
                signal_builder_wrap.signal_bundle_builder.reset()
                signal_builder_wrap.is_being_emitted = False

    async def _schedule_signal_auto_emit(self, wrapper_key, delay):
        while wrapper_key in self._signal_builder_wrappers:
            await asyncio.sleep(delay)
            if wrapper_key not in self._signal_builder_wrappers:
                # signal should not be emitted anymore
                break
            wrapper = self._signal_builder_wrappers[wrapper_key]
            if time.time() >= wrapper.signal_emit_time:
                await self._emit_signal_if_necessary(wrapper)
                wrapper.signal_emit_time = time.time() + delay

    def _register_timeout_if_any(self, wrapper_key):
        wrapper = self._signal_builder_wrappers[wrapper_key]
        if wrapper.timeout != wrapper.NO_TIMEOUT_VALUE:
            self._timeout_watcher_tasks[wrapper_key] = asyncio.create_task(
                self._schedule_signal_auto_emit(wrapper_key, wrapper.timeout)
            )

    def _unregister_timeout(self, wrapper_key):
        if task := self._timeout_watcher_tasks.pop(wrapper_key, None):
            task.cancel()



# pylint: disable=R0801
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_commons.dsl_interpreter.operators.binary_operator import (
    BinaryOperator,
)
from octobot_commons.dsl_interpreter.operators.compare_operator import (
    CompareOperator,
)
from octobot_commons.dsl_interpreter.operators.unary_operator import (
    UnaryOperator,
)
from octobot_commons.dsl_interpreter.operators.n_ary_operator import (
    NaryOperator,
)
from octobot_commons.dsl_interpreter.operators.call_operator import (
    CallOperator,
)
from octobot_commons.dsl_interpreter.operators.name_operator import (
    NameOperator,
)
from octobot_commons.dsl_interpreter.operators.expression_operator import (
    ExpressionOperator,
)
from octobot_commons.dsl_interpreter.operators.subscripting_operator import (
    SubscriptingOperator,
)
from octobot_commons.dsl_interpreter.operators.iterable_operator import (
    IterableOperator,
)

__all__ = [
    "BinaryOperator",
    "CompareOperator",
    "UnaryOperator",
    "NaryOperator",
    "CallOperator",
    "NameOperator",
    "ExpressionOperator",
    "SubscriptingOperator",
    "IterableOperator",
]



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_commons.context_util as context_util


def test_empty_context_manager_sync_enter_exit():
    """Test EmptyContextManager as synchronous context manager"""
    manager = context_util.EmptyContextManager()
    
    # Test that __enter__ returns self
    with manager as ctx:
        assert ctx is manager
        # Verify we can use it normally
        assert ctx is not None


def test_empty_context_manager_sync_no_exception_suppression():
    """Test that EmptyContextManager doesn't suppress exceptions"""
    manager = context_util.EmptyContextManager()
    
    # Exception should propagate normally
    with pytest.raises(ValueError):
        with manager:
            raise ValueError("Test exception")


def test_empty_context_manager_sync_multiple_usage():
    """Test EmptyContextManager can be used multiple times"""
    manager = context_util.EmptyContextManager()
    
    # First usage
    with manager as ctx1:
        assert ctx1 is manager
    
    # Second usage
    with manager as ctx2:
        assert ctx2 is manager


def test_empty_context_manager_sync_exit_with_exception():
    """Test EmptyContextManager __exit__ with exception"""
    manager = context_util.EmptyContextManager()
    
    # __exit__ should not suppress exceptions (returns None/False)
    with pytest.raises(ValueError):
        with manager:
            raise ValueError("Test exception")
    
    # Verify exception was not suppressed
    try:
        with manager:
            raise ValueError("Another exception")
    except ValueError as e:
        assert str(e) == "Another exception"


def test_empty_context_manager_sync_exit_without_exception():
    """Test EmptyContextManager __exit__ without exception"""
    manager = context_util.EmptyContextManager()
    
    # Normal exit should work fine
    with manager:
        pass
    
    # Should complete without issues
    assert True


@pytest.mark.asyncio
async def test_empty_context_manager_async_enter_exit():
    """Test EmptyContextManager as asynchronous context manager"""
    manager = context_util.EmptyContextManager()
    
    # Test that __aenter__ returns self
    async with manager as ctx:
        assert ctx is manager
        # Verify we can use it normally
        assert ctx is not None


@pytest.mark.asyncio
async def test_empty_context_manager_async_no_exception_suppression():
    """Test that EmptyContextManager doesn't suppress exceptions in async context"""
    manager = context_util.EmptyContextManager()
    
    # Exception should propagate normally
    with pytest.raises(ValueError):
        async with manager:
            raise ValueError("Test async exception")


@pytest.mark.asyncio
async def test_empty_context_manager_async_multiple_usage():
    """Test EmptyContextManager can be used multiple times in async context"""
    manager = context_util.EmptyContextManager()
    
    # First usage
    async with manager as ctx1:
        assert ctx1 is manager
    
    # Second usage
    async with manager as ctx2:
        assert ctx2 is manager


@pytest.mark.asyncio
async def test_empty_context_manager_async_exit_with_exception():
    """Test EmptyContextManager __aexit__ with exception"""
    manager = context_util.EmptyContextManager()
    
    # __aexit__ should not suppress exceptions (returns None/False)
    with pytest.raises(ValueError):
        async with manager:
            raise ValueError("Test async exception")
    
    # Verify exception was not suppressed
    try:
        async with manager:
            raise ValueError("Another async exception")
    except ValueError as e:
        assert str(e) == "Another async exception"


@pytest.mark.asyncio
async def test_empty_context_manager_async_exit_without_exception():
    """Test EmptyContextManager __aexit__ without exception"""
    manager = context_util.EmptyContextManager()
    
    # Normal exit should work fine
    async with manager:
        pass
    
    # Should complete without issues
    assert True


@pytest.mark.asyncio
async def test_empty_context_manager_mixed_sync_async():
    """Test EmptyContextManager can be used both synchronously and asynchronously"""
    manager = context_util.EmptyContextManager()
    
    # Use synchronously
    with manager:
        pass
    
    # Use asynchronously
    async with manager:
        pass
    
    # Both should work without issues
    assert True



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import typing


@dataclasses.dataclass
class FlexibleDataclass:
    _class_field_cache: typing.ClassVar[dict] = {}
    """
    Implements from_dict which can be called to instantiate a new instance of this class from a dict. Using from_dict 
    ignores any additional key from the given dict that is not defined as a dataclass field.
    Nested dataclasses to be parsed inside a list or other container should be calling .from_dict in __post_init__
    """

    @classmethod
    def from_dict(cls, dict_value: dict):
        """
        Creates a new instance of cls from the given dict, ignoring additional dict values
        """
        if isinstance(dict_value, dict):
            fields_values = {
                k: _get_nested_class(v, cls._class_field_cache[k])
                for k, v in dict_value.items()
                if k in cls.get_field_names()
            }
            try:
                return cls(**fields_values)
            except TypeError as e:
                raise TypeError(
                    f"Invalid {cls.__name__} input in from_dict(): {e}"
                ) from e
        return dict_value

    @classmethod
    def get_field_names(cls):
        """
        :return a generator over the given FlexibleDataclass field names
        """
        if not cls._class_field_cache:
            cls._class_field_cache = {
                f.name: f.type for f in dataclasses.fields(cls) if f.init
            }
        return cls._class_field_cache.keys()


def _get_nested_class(value, target_type):
    # does not support lists or dicts
    if isinstance(target_type, type) and issubclass(target_type, FlexibleDataclass):
        return target_type.from_dict(value)
    return value



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


import sys
import os
import platform
import ctypes

import octobot_commons.constants as constants
import octobot_commons.enums as enums

try:
    import psutil
except ImportError:
    if constants.USE_MINIMAL_LIBS:
        # mock psutil imports
        class PsutilImportMock:
            class virtual_memory:  # pylint: disable=invalid-name
                def __init__(self, *args):
                    raise ImportError("psutil not installed")

                def __getitem__(self, key):
                    return self[key]

            class Process:
                def __init__(self, *args):
                    raise ImportError("psutil not installed")

                def memory_info(self):  # pylint: disable=missing-function-docstring
                    class MemoryInfo:
                        def __init__(self, *args):
                            pass

                        def rss(self):  # pylint: disable=missing-function-docstring
                            return 0

                    return MemoryInfo()

            class cpu_percent:  # pylint: disable=invalid-name
                def __init__(self, *args):
                    raise ImportError("psutil not installed")

        psutil = PsutilImportMock()
    else:
        raise


def get_current_platform():
    """
    Return the current platform details
    Return examples
    For Windows :
    >>> 'Windows:10:AMD64'
    For Linux :
    >>> 'Linux:4.15.0-46-generic:x86_64'
    For Raspberry :
    >>> 'Linux:4.14.98-v7+:armv7l'
    :return: the current platform details
    """
    return (
        f"{platform.system()}{constants.PLATFORM_DATA_SEPARATOR}{platform.release()}{constants.PLATFORM_DATA_SEPARATOR}"
        f"{platform.machine()}"
    )


def get_octobot_type():
    """
    Return OctoBot running type from OctoBotTypes
    :return: the OctoBot running type
    """
    try:
        execution_arg = sys.argv[0]
        # sys.argv[0] is always the name of the python script called when using a command "python xyz.py"
        if execution_arg.endswith(".py"):
            if _is_on_docker():
                return enums.OctoBotTypes.DOCKER.value
            return enums.OctoBotTypes.PYTHON.value
        # sys.argv[0] is the name of the binary when using a binary version: ends with nothing or .exe"
        return enums.OctoBotTypes.BINARY.value
    except IndexError:
        return enums.OctoBotTypes.BINARY.value


def get_os():
    """
    Return the OS name
    :return: the OS name
    """
    return enums.PlatformsName(os.name)


def has_admin_rights() -> bool:
    """
    :return: True if the current thread has admin rights
    """
    try:
        return os.getuid() == 0
    except AttributeError:
        return ctypes.windll.shell32.IsUserAnAdmin()


def is_machine_64bit() -> bool:
    """
    Win: AMD64
    Debian-64: x86_64
    From https://stackoverflow.com/questions/2208828/detect-64bit-os-windows-in-python
    :return: True if the machine is 64bit
    """
    return platform.machine().endswith("64")


def is_arm_machine() -> bool:
    """
    Can be armv7l or aarch64 (raspberry, Android smartphone...)
    From https://raspberrypi.stackexchange.com/questions/5100/detect-that-a-python-program-is-running-on-the-pi
    :return: True if the machine is 64bit
    """
    return platform.machine() in ["armv7l", "aarch64"]


def is_raspberry_pi_machine() -> bool:
    """
    Check if the machine is a Raspberry Pi
    Works on bare metal and in Docker containers
    :return: True if the machine is a Raspberry Pi
    """
    # Try device-tree method (works on bare metal Raspberry Pi)
    try:
        with open("/proc/device-tree/model", "r") as f:
            if "Raspberry Pi" in f.read():
                return True
    except (FileNotFoundError, IOError):
        pass

    # Fallback to cpuinfo (works in Docker containers)
    try:
        with open("/proc/cpuinfo", "r") as f:
            cpuinfo = f.read()
            return "Raspberry Pi" in cpuinfo or "BCM" in cpuinfo
    except (FileNotFoundError, IOError):
        pass
    return False


def _is_on_docker():
    """
    Check if the current platform is docker
    :return: True if OctoBot is running with docker
    """
    file_to_check = "/proc/self/cgroup"
    try:
        return os.path.exists("/.dockerenv") or (
            os.path.isfile(file_to_check)
            and any("docker" in line for line in open(file_to_check))
        )
    except FileNotFoundError:
        return False


def parse_boolean_environment_var(env_key: str, default_value: str) -> bool:
    """
    Parse a boolean environment variable
    :param env_key: the environment variable key
    :param default_value: the default value
    :return: the boolean value
    """
    return constants.parse_boolean_environment_var(env_key, default_value)


def get_cpu_and_ram_usage(cpu_watching_seconds):
    """
    WARNING: blocking the current thread for the given cpu_watching_seconds seconds
    :return: the CPU usage percent, RAM usaage %, total RAM used and total RAM used by this process
    """
    mem_ret = psutil.virtual_memory()
    process_used_ram = psutil.Process(os.getpid()).memory_info().rss
    return (
        psutil.cpu_percent(cpu_watching_seconds),
        mem_ret[2],
        mem_ret[3] / constants.BYTES_BY_GB,
        process_used_ram / constants.BYTES_BY_GB,
    )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import inspect

import octobot_commons.logging as logging_util


def default_parent_inspection(element, parent):
    """
    Check if the element bases has the specified parent
    :param element: the element to check
    :param parent: the expected parent
    :return: the check result
    """
    return parent in element.__bases__


def default_parents_inspection(element, parent):
    """
    Check if the element has the specified parent
    :param element: the element to check
    :param parent: the expected parent
    :return: the check result
    """
    return parent in element.mro()


def evaluator_parent_inspection(element, parent):
    """
    Recursively check if the evaluator class has the specified parent
    :param element: the element to check
    :param parent: the expected parent
    :return: the check result
    """
    return hasattr(
        element, "get_parent_evaluator_classes"
    ) and element.get_parent_evaluator_classes(parent)


def trading_mode_parent_inspection(element, parent):
    """
    Check if the trading class has the specified parent
    :param element: the element to check
    :param parent: the expected parent
    :return: the check result
    """
    return hasattr(
        element, "get_parent_trading_mode_classes"
    ) and element.get_parent_trading_mode_classes(parent)


def get_class_from_parent_subclasses(class_string, parent):
    """
    Search the class string in parent subclasses
    :param class_string: the class name to search
    :param parent: the parent
    :return: the class if found else None
    """
    for found in parent.__subclasses__():
        if found.__name__ == class_string:
            return found
    return None


def get_deep_class_from_parent_subclasses(class_string, parent):
    """
    Search for a class in parent subclasses "deeply"
    :param class_string: the class name to search
    :param parent: the expected parent
    :return: the class if found else None
    """
    found = get_class_from_parent_subclasses(class_string, parent)
    if found is not None:
        return found

    for parent_class in parent.__subclasses__():
        found = get_deep_class_from_parent_subclasses(class_string, parent_class)
        if found is not None:
            return found
    return None


def get_class_from_string(
    class_string: str,
    parent,
    module,
    parent_inspection=default_parent_inspection,
    error_when_not_found: bool = False,
):
    """
    Search a class from a class string in a specified module for a specified parent
    :param class_string: the class name to search
    :param parent: the class expected parent
    :param module: the class expected module
    :param parent_inspection: the parent inspection
    :param error_when_not_found: if errors should be raised
    :return: the class if found else None
    """
    if tentacle_class_by_name := {
        m[0]: m[1]
        for m in inspect.getmembers(module)
        if (m[0] == class_string)
        and hasattr(m[1], "__bases__")
        and parent_inspection(m[1], parent)
    }:
        return tentacle_class_by_name[class_string]
    if error_when_not_found:
        raise ModuleNotFoundError(f"Cant find {class_string} module")
    return None  # no class found


def is_abstract_using_inspection_and_class_naming(clazz):
    """
    Check if a class is abstract
    :param clazz: the class to check
    :return: the check result
    """
    return inspect.isabstract(clazz) or "abstract" in clazz.__name__.lower()


def get_all_classes_from_parent(parent_class) -> list:
    """
    Get all sub classes from parent including multi level sub-classes
    :param parent_class: the parent class
    :return: the class from parent
    """
    classes = []
    for subclass in parent_class.__subclasses__():
        if subclass.__subclasses__():
            # append this subclass
            classes.append(subclass)

            # and all its subclasses
            classes += get_all_classes_from_parent(subclass)
        else:
            classes.append(subclass)
    return classes


def get_single_deepest_child_class(clazz) -> object:
    """
    Get the single deepest child class
    :param clazz: the class
    :return: the single deepest child class
    """
    children_classes = clazz.__subclasses__()
    if len(children_classes) == 0:
        return clazz
    if len(children_classes) > 1:
        logging_util.get_logger(__name__).error(
            f"More than one child class of {clazz}, expecting one, "
            f"using {children_classes[0]}"
        )
    return get_single_deepest_child_class(children_classes[0])



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.configuration as configuration
from octobot_commons.tentacles_management.abstract_tentacle import AbstractTentacle


class TentacleTest(AbstractTentacle):
    def __init__(self):
        super().__init__()


class TentacleTestChild(TentacleTest):
    def __init__(self):
        super().__init__()
        self.plop = 1


def test_get_name():
    assert TentacleTest().get_name() == "TentacleTest"
    assert TentacleTestChild().get_name() == "TentacleTestChild"


def test_get_all_subclasses():
    assert TentacleTest().get_all_subclasses() == [TentacleTestChild]


def test_user_input_factories():
    tentacle = TentacleTest()
    assert isinstance(tentacle.UI, configuration.UserInputFactory)
    assert isinstance(tentacle.CLASS_UI, configuration.UserInputFactory)
    assert isinstance(tentacle.__class__.CLASS_UI, configuration.UserInputFactory)
    assert isinstance(TentacleTestChild.CLASS_UI, configuration.UserInputFactory)
    assert TentacleTestChild.CLASS_UI is tentacle.CLASS_UI

    child_tentacle = TentacleTestChild()
    assert TentacleTestChild.CLASS_UI is child_tentacle.CLASS_UI
    assert child_tentacle.plop == 1
    assert isinstance(child_tentacle.UI, configuration.UserInputFactory)
    assert isinstance(child_tentacle.CLASS_UI, configuration.UserInputFactory)
    assert isinstance(tentacle.__class__.CLASS_UI, configuration.UserInputFactory)
    assert isinstance(TentacleTestChild.CLASS_UI, configuration.UserInputFactory)
    assert TentacleTestChild.CLASS_UI is child_tentacle.CLASS_UI
    assert tentacle.CLASS_UI is child_tentacle.CLASS_UI



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import copy

import octobot_commons.enums
import octobot_commons.signals as signals


@pytest.fixture
def no_dependencies():
    return signals.SignalDependencies()


@pytest.fixture
def single_dependencies():
    return signals.SignalDependencies([
        {"PLIP": "123"}
    ])


@pytest.fixture
def dual_dependencies():
    return signals.SignalDependencies([
        {"plop": "123"},
        {"PLIP": "123"}
    ])


def test_extend(no_dependencies, single_dependencies, dual_dependencies):
    no_dependencies.extend(no_dependencies)
    assert no_dependencies == no_dependencies
    no_dependencies.extend(single_dependencies)
    assert no_dependencies == single_dependencies
    single_dependencies.extend(dual_dependencies)
    assert single_dependencies == signals.SignalDependencies([
        {"PLIP": "123"},
        {"plop": "123"},
        {"PLIP": "123"}
    ])


def test_is_filled_by(no_dependencies, single_dependencies, dual_dependencies):
    assert no_dependencies.is_filled_by(no_dependencies) is True
    assert no_dependencies.is_filled_by(single_dependencies) is True
    assert no_dependencies.is_filled_by(dual_dependencies) is True


    assert single_dependencies.is_filled_by(no_dependencies) is False
    assert single_dependencies.is_filled_by(single_dependencies) is True
    assert single_dependencies.is_filled_by(dual_dependencies) is True

    assert dual_dependencies.is_filled_by(no_dependencies) is False
    assert dual_dependencies.is_filled_by(single_dependencies) is False
    assert dual_dependencies.is_filled_by(dual_dependencies) is True

    # with extended dependencies
    saved_no_dependencies = copy.deepcopy(no_dependencies)
    no_dependencies.extend(single_dependencies)
    assert no_dependencies.is_filled_by(saved_no_dependencies) is False
    assert no_dependencies.is_filled_by(single_dependencies) is True
    assert no_dependencies.is_filled_by(dual_dependencies) is True


def test_to_dict(no_dependencies, single_dependencies, dual_dependencies):
    assert no_dependencies.to_dict() == {
        octobot_commons.enums.SignalDependenciesAttrs.DEPENDENCY.value: []
    }
    assert single_dependencies.to_dict() == {
        octobot_commons.enums.SignalDependenciesAttrs.DEPENDENCY.value: [
            {"PLIP": "123"}
        ]
    }
    assert dual_dependencies.to_dict() == {
        octobot_commons.enums.SignalDependenciesAttrs.DEPENDENCY.value: [
            {"plop": "123"},
            {"PLIP": "123"}
        ]
    }


def test_eq(no_dependencies, single_dependencies, dual_dependencies):
    assert no_dependencies != None
    assert no_dependencies == no_dependencies
    assert no_dependencies == signals.SignalDependencies()
    assert single_dependencies == single_dependencies
    assert single_dependencies == signals.SignalDependencies([
        {"PLIP": "123"}
    ])
    assert dual_dependencies == dual_dependencies
    assert dual_dependencies == signals.SignalDependencies([
        {"plop": "123"},
        {"PLIP": "123"}
    ])
    assert no_dependencies != single_dependencies


def test_bool(no_dependencies, single_dependencies, dual_dependencies):
    assert bool(no_dependencies) is False
    assert bool(single_dependencies) is True
    assert bool(dual_dependencies) is True



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_commons.enums import TimeFrames
from octobot_commons.tests.test_config import load_test_config
from octobot_commons.time_frame_manager import get_config_time_frame, parse_time_frames, find_min_time_frame, \
    get_previous_time_frame, get_display_time_frame, sort_time_frames


def test_get_config_time_frame():
    assert get_config_time_frame(load_test_config()) == [TimeFrames("1h"), TimeFrames("4h"), TimeFrames("1d")]


def test_parse_time_frames():
    assert parse_time_frames(["3d", "5d", "1m", "6h"]) == [TimeFrames("3d"), TimeFrames("1m"), TimeFrames("6h")]


def test_sort_time_frames():
    assert sort_time_frames([TimeFrames("3d"), TimeFrames("1m"), TimeFrames("6h")]) == \
           [TimeFrames("1m"), TimeFrames("6h"), TimeFrames("3d")]
    assert sort_time_frames([TimeFrames("1M"), TimeFrames("3d"), TimeFrames("12h"), TimeFrames("1h"),
                             TimeFrames("1m"), TimeFrames("6h")]) == \
           [TimeFrames("1m"), TimeFrames("1h"), TimeFrames("6h"), TimeFrames("12h"), TimeFrames("3d"), TimeFrames("1M")]


def test_find_min_time_frame():
    assert find_min_time_frame([TimeFrames.FOUR_HOURS, TimeFrames.ONE_DAY, TimeFrames.ONE_MONTH,
                                TimeFrames.FIFTEEN_MINUTES]) == TimeFrames.FIFTEEN_MINUTES
    assert find_min_time_frame([TimeFrames.ONE_MONTH, TimeFrames.ONE_WEEK]) == TimeFrames.ONE_WEEK
    assert find_min_time_frame([TimeFrames.ONE_MINUTE]) == TimeFrames.ONE_MINUTE


def test_get_previous_time_frame():
    assert get_previous_time_frame(get_config_time_frame(load_test_config()),
                                   TimeFrames.ONE_DAY, TimeFrames.ONE_DAY) == TimeFrames.FOUR_HOURS
    assert get_previous_time_frame(get_config_time_frame(load_test_config()),
                                   TimeFrames.ONE_MINUTE, TimeFrames.ONE_MINUTE) == TimeFrames.ONE_MINUTE
    assert get_previous_time_frame(get_config_time_frame(load_test_config()),
                                   TimeFrames.ONE_HOUR, TimeFrames.ONE_HOUR) == TimeFrames.ONE_HOUR
    assert get_previous_time_frame(get_config_time_frame(load_test_config()),
                                   TimeFrames.ONE_MONTH, TimeFrames.ONE_MONTH) == TimeFrames.ONE_DAY


def test_get_display_time_frame():
    assert get_display_time_frame(load_test_config(), TimeFrames.ONE_MONTH) == TimeFrames.ONE_DAY
    assert get_display_time_frame(load_test_config(), TimeFrames.FOUR_HOURS) == TimeFrames.FOUR_HOURS
